#! /usr/bin/env python
"""Implements the functionality in MESA's EGL + OpenGL demo1:

    http://cgit.freedesktop.org/mesa/demos/tree/src/egl/opengl/demo1.c

Basically just to check if the EGL code works...

NOTE: this test program is both MIT (MESA) and GPLv2 licensed due to use of 
sample code from python-Xlib.
"""
import pygame.display 
import pygame 
import os 
import logging
import numpy
if not os.environ.get( 'PYOPENGL_PLATFORM' ):
    os.environ['PYOPENGL_PLATFORM'] = 'egl'

import OpenGL,ctypes
if os.environ.get( 'TEST_NO_ACCELERATE' ):
    OpenGL.USE_ACCELERATE = False
from OpenGL._bytes import as_str
from OpenGL.EGL import *
from OpenGL.error import GLError
log = logging.getLogger( __name__ )
import sys
from OpenGL.arrays.vbo import VBO

def describe_config( display, config ):
    """Describe the given configuration"""
    parameters = (EGL_CONFIG_ID,
        EGL_BUFFER_SIZE,
        EGL_LEVEL,
        EGL_RED_SIZE,
        EGL_GREEN_SIZE,
        EGL_BLUE_SIZE,
        EGL_ALPHA_SIZE,
        EGL_DEPTH_SIZE,
        EGL_STENCIL_SIZE,
        EGL_SURFACE_TYPE)
    description = []
    for param in parameters:
        value = ctypes.c_long()
        eglGetConfigAttrib(display, config, param, value)
        description.append( '%s = %s'%( param, value.value, ))
    return '\n'.join( description )

def mainloop( displayfunc ):
    while True:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                return 
        displayfunc()
    

def main(displayfunc, api):
    major,minor = ctypes.c_long(),ctypes.c_long()
    display = eglGetDisplay(EGL_DEFAULT_DISPLAY)
    log.info( 'Display return value: %s', display )
    log.info( 'Display address: %s', display.address)
    #display = display.as_voidp
    #print 'wrapped', display
    if not eglInitialize( display, major, minor):
        log.error( 'Unable to initialize' )
        return 
    log.info( 'EGL version %s.%s', major.value,minor.value)
    
    num_configs = ctypes.c_long()
    eglGetConfigs(display, None, 0, num_configs)
    log.info( '%s configs', num_configs.value)
    
    configs = (EGLConfig * num_configs.value)()
    eglGetConfigs(display,configs,num_configs.value,num_configs)
    for config_id in configs:
        #print config_id
        describe_config( display, config_id )
    
    log.info( 'Attempting to bind and create contexts/apis' )
    try:
        eglBindAPI( 0x3333 ) # junk value 
    except GLError:
        pass 
    else:
        assert False, "Should have generated error on bind to non-existent api"
    eglBindAPI(api)
    
    # now need to get a raw X window handle...
    pygame.init()
    pygame.display.set_mode( (500,500) )
    window = pygame.display.get_wm_info()['window']
    surface = eglCreateWindowSurface(display, configs[0], window, None )
    
    ctx = eglCreateContext(display, configs[0], EGL_NO_CONTEXT, None)
    if ctx == EGL_NO_CONTEXT:
        log.error( 'Unable to create the regular context' )
        return 
    else:
        log.info( 'Created regular context' )
        def _displayfunc( ):
            displayfunc( display, surface, ctx )
        mainloop( _displayfunc )
    
    pbufAttribs = (EGLint * 5)(* [EGL_WIDTH,500, EGL_HEIGHT, 500, EGL_NONE])
    pbuffer = eglCreatePbufferSurface(display, configs[0], pbufAttribs);
    if (pbuffer == EGL_NO_SURFACE):
        log.error( 'Unable to create pbuffer surface' )
    else:
        log.info( 'created pbuffer surface' )
    log.info( 'Available EGL extensions:\n  %s', "\n  ".join([
        as_str(ext) for ext in 
        EGLQuerier.getExtensions().split()
    ]))

def displayfunc_gl( display, surface, ctx ):
    from OpenGL import GL
    eglMakeCurrent( display, surface, surface, ctx )
    GL.glClearColor( 1,0,0, 0 )
    GL.glClear( GL.GL_COLOR_BUFFER_BIT|GL.GL_DEPTH_BUFFER_BIT )
    eglSwapBuffers( display, surface )
displayfunc_gl.api = EGL_OPENGL_API

def displayfunc_gles1(display, surface, ctx):
    from OpenGL import GLES1 as GL
    eglMakeCurrent( display, surface, surface, ctx )
    GL.glClearColor( 1,0,0, 0 )
    GL.glClear( GL.GL_COLOR_BUFFER_BIT|GL.GL_DEPTH_BUFFER_BIT )
    vertices = numpy.array( ( (1,0,0 ),(-1,0,0 ),(0,1,0 )), 'f')
    GL.glEnableClientState( GL.GL_VERTEX_ARRAY )
    GL.glVertexPointer( 3, GL.GL_FLOAT, 0, vertices )
    GL.glDrawArrays( GL.GL_TRIANGLES, 0, 3 )
    eglSwapBuffers( display, surface )
displayfunc_gles1.api = EGL_OPENGL_ES_API


shader = None
def displayfunc_gles2(display, surface, ctx):
    from OpenGL import GLES2 as GL
    from OpenGL.GLES2 import shaders
    eglMakeCurrent( display, surface, surface, ctx )
    GL.glClearColor( 1,0,0, 0 )
    GL.glClear( GL.GL_COLOR_BUFFER_BIT|GL.GL_DEPTH_BUFFER_BIT )
    
    global shader,vbo,position_location,stride
    if shader is None:
        shader = shaders.compileProgram(
            shaders.compileShader( '''#version 130
    attribute vec3 position;
    void main() {
        gl_Position = vec4( position, 0 );
    }''', type=GL_VERTEX_SHADER),
            shaders.compileShader( '''#version 130
    void main() {
        gl_FragColor = vec4( 1,0,0,0 );
    }''', type=GL_FRAGMENT_SHADER),
        )
        vbo = VBO( array([
            (0,1,0),
            (1,-1,0),
            (-1,-1,0),
        ],dtype='f'))
        position_location = GL.glGetAttribLocation(
            self.shader, 'position'
        )
        stride = 3*4
    with vbo:
        with shader:
            GL.glEnableVertexAttribArray( position_location )
            stride = 3*4
            GL.glVertexAttribPointer(
                position_location,
                3, GL_FLOAT,False, stride, self.vbo
            )
            GL.glDrawArrays( GL.GL_TRIANGLES, 0, 3 )
    eglSwapBuffers( display, surface )
    
displayfunc_gles2.api = EGL_OPENGL_ES_API

def displayfunc_gles3(display, surface, ctx):
    from OpenGL import GLES3 as GL
    eglMakeCurrent( display, surface, surface, ctx )
    GL.glClearColor( 1,0,0, 0 )
    GL.glClear( GL.GL_COLOR_BUFFER_BIT|GL.GL_DEPTH_BUFFER_BIT )
    eglSwapBuffers( display, surface )
displayfunc_gles3.api = EGL_OPENGL_ES_API

    
if __name__ == "__main__":
    logging.basicConfig( level=logging.INFO )
    if sys.argv[1:]:
        name = sys.argv[1]
    else:
        name = 'gl'
    if name == 'gles':
        name = 'gles1'
    function = globals().get( 'displayfunc_%s'%(name,), displayfunc_gl )
    log.info( 'Using function: %s', function )
    main(function,function.api)



import pygamegltest
import os

# We have to import at least *one* VBO implementation...
from OpenGL import GL, arrays
from OpenGL.arrays import vbo

try:
    import psutil
except ImportError:
    psutil = None
try:
    unicode
except NameError:
    unicode = str
    long = int
import pytest
import gc

try:
    import numpy as np
except ImportError:
    np = None


def get_current_memory():
    return psutil.Process(os.getpid()).memory_info().rss


@pytest.mark.skipif(not psutil, reason="No psutil available")
@pytest.mark.skipif(not np, reason="No Numpy available")
@pygamegltest.pygametest()
def test_sf_2980896():
    """Test SF#2980896 report of memory leak on VBO transfer"""
    data = arrays.GLfloatArray.zeros((1000,))
    memory = get_current_memory()
    for i in range(100):
        new_vbo = vbo.VBO(data)
        with new_vbo:
            # data is transferred to the VBO
            assert new_vbo is not None, new_vbo
        new_vbo.delete()
        del new_vbo
        gc.collect()
        GL.glFinish()
        if i < 1:
            # the *first* call can load lots of libraries, etc...
            memory = get_current_memory()
        else:
            current = get_current_memory()
            assert current - memory < 200, (
                """Shouldn't have any (or at least much) extra RAM allocated, lost: %s"""
                % (current - memory)
            )  # fails only when run in the whole suite...



"""Transliteration of https://open.gl/feedback into Python"""
from __future__ import print_function
import pygamegltest
from OpenGL.GL import *
from OpenGL.GL import shaders
vertex_shader = """#version 150 core
    in float inValue;
    out float outValue;

    void main() {
        outValue = sqrt(inValue);
    }
"""

@pygamegltest.pygametest(name="Geometry Shader Test")
def main():
    program = shaders.compileProgram(
        shaders.compileShader([vertex_shader], GL_VERTEX_SHADER), 
    )

    buff = ctypes.c_char_p( b"outValue" )
    c_text = ctypes.cast(ctypes.pointer(buff), ctypes.POINTER(ctypes.POINTER(GLchar)))
    # modifies the state in the linking of the program
    glTransformFeedbackVaryings(program, 1, c_text, GL_INTERLEAVED_ATTRIBS);
    
    # so have to re-link
    glLinkProgram(program)
    glUseProgram(program);

    vao = glGenVertexArrays(1);
    glBindVertexArray(vao);

    data = (GLfloat * 5)(1.0, 2.0, 3.0, 4.0, 5.0)
    
    vbo = glGenBuffers(1);
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    glBufferData(GL_ARRAY_BUFFER, ctypes.sizeof(data), data, GL_STATIC_DRAW);

    inputAttrib = glGetAttribLocation(program, "inValue");
    glEnableVertexAttribArray(inputAttrib);
    # Note the need to cast 0 to a GLvoidp here!
    glVertexAttribPointer(inputAttrib, 1, GL_FLOAT, GL_FALSE, 0, GLvoidp(0));

    tbo = glGenBuffers(1);
    glBindBuffer(GL_ARRAY_BUFFER, tbo);
    glBufferData(GL_ARRAY_BUFFER, ctypes.sizeof(data), None, GL_STATIC_READ);

    glEnable(GL_RASTERIZER_DISCARD);

    glBindBufferBase(GL_TRANSFORM_FEEDBACK_BUFFER, 0, tbo);

    glBeginTransformFeedback(GL_POINTS);
    glDrawArrays(GL_POINTS, 0, len(data));
    glEndTransformFeedback();

    glDisable(GL_RASTERIZER_DISCARD);

    glFlush();

    feedback = (GLfloat*len(data))()
    glGetBufferSubData(GL_TRANSFORM_FEEDBACK_BUFFER, 0, ctypes.sizeof(feedback), feedback);

    for value in feedback:
        print(value)

if __name__ == "__main__":
    main()



"""Test GLUT forward-compatible mode..."""
from __future__ import print_function
import OpenGL
OpenGL.FORWARD_COMPATIBLE_ONLY = True
OpenGL.ERROR_CHECKING = True
#OpenGL.USE_ACCELERATE = False
from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *
import time
start = time.time()

from OpenGL.GL.AMD.debug_output import glGetDebugMessageLogAMD
from OpenGL.GL.ARB.debug_output import glGetDebugMessageLogARB
from OpenGL.GL.KHR.debug import glGetDebugMessageLogKHR
from OpenGL.GL.VERSION.GL_4_3 import glGetDebugMessageLog

glGetDebugMessageLog = extensions.alternate(
    glGetDebugMessageLog,
    glGetDebugMessageLogARB,
    glGetDebugMessageLogKHR,
    glGetDebugMessageLogAMD,
)

window = None

def display():
    try:
        glutSetWindow(window);
        glClearColor (0.0, 0.0, (time.time()%1.0)/1.0, 0.0)
        glClear (GL_COLOR_BUFFER_BIT)
        try:
            glGetString( GL_EXTENSIONS )
        except GLError:
            pass 
        else:
            print('Egads, glGetString should not have worked!')
        assert bool( glGenVertexArrays ), "Should have vertex array support in 3.2"
        for message in get_debug_messages():
            print(message)
        glFlush ()
        glutSwapBuffers()
    except Exception:
        glutDestroyWindow( window )
        raise

def get_debug_messages():
    messages = []
    count = glGetIntegerv( GL_DEBUG_LOGGED_MESSAGES )
    max_size = int(glGetIntegerv( GL_MAX_DEBUG_MESSAGE_LENGTH ))
    source = GLenum()
    type = GLenum()
    id = GLenum()
    severity = GLenum()
    length = GLsizei()
    buffer = ctypes.create_string_buffer( max_size )
    for i in range(count):
        result = glGetDebugMessageLog( 1, max_size, source, type, id, severity, length, buffer )
        if result:
            messages.append( {
                'message':buffer[:length.value],
                'type': type.value,
                'id': id.value,
                'severity': severity.value,
                'source': source.value,
            })
    return messages
    assert len(messages), messages

        

size = (250,250)

def reshape( *args ):
    global size 
    size = args
    glViewport( *( (0,0)+args) )
    display()

def printFunction( name ):
    def onevent( *args ):
        print('%s -> %s'%(name, ", ".join( [str(a) for a in args ])))
    return onevent



if __name__ == "__main__":
    import sys
    newArgv = glutInit(sys.argv)
    glutInitContextVersion(3, 1)
    glutInitContextFlags(GLUT_FORWARD_COMPATIBLE|GLUT_DEBUG)
    glutInitContextProfile(GLUT_CORE_PROFILE)
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH )

    glutSetOption(
        GLUT_ACTION_ON_WINDOW_CLOSE,
        GLUT_ACTION_GLUTMAINLOOP_RETURNS
    );
    glutInitWindowSize(250, 250)
    glutInitWindowPosition(100, 100)
    window = glutCreateWindow("hello")
    print('window', repr(window))
    glutDisplayFunc(display)
    glutReshapeFunc(reshape)
    glutMouseFunc(printFunction( 'Mouse' ))
    glutEntryFunc(printFunction( 'Entry' ))
    glutKeyboardFunc( printFunction( 'Keyboard' ))
    glutKeyboardUpFunc( printFunction( 'KeyboardUp' ))
    glutMotionFunc( printFunction( 'Motion' ))
    glutPassiveMotionFunc( printFunction( 'PassiveMotion' ))
    glutVisibilityFunc( printFunction( 'Visibility' ))
    glutWindowStatusFunc( printFunction( 'WindowStatus' ))
    glutSpecialFunc( printFunction( 'Special' ))
    glutSpecialUpFunc( printFunction( 'SpecialUp' ))
    
    glutMainLoop()



#! /usr/bin/env python
import egltest
import numpy
from OpenGL import GLES1 as GL

@egltest.egltest( api='gles1' )
def test_es1( ):
    GL.glClearColor( 1,0,0, 0 )
    GL.glClear( GL.GL_COLOR_BUFFER_BIT|GL.GL_DEPTH_BUFFER_BIT )
    vertices = numpy.array( ( (1,0,0 ),(-1,0,0 ),(0,1,0 )), 'f')
    GL.glEnableClientState( GL.GL_VERTEX_ARRAY )
    GL.glVertexPointer( 3, GL.GL_FLOAT, 0, vertices )
    GL.glDrawArrays( GL.GL_TRIANGLES, 0, 3 )

if __name__ == "__main__":
    test_es1()
    



import basetestcase
import os, sys
import OpenGL
from OpenGL.arrays import (
    arraydatatype
)
from OpenGL.GL import *
try:
    import numpy as np
except ImportError as err:
    np = None
HERE = os.path.abspath(os.path.dirname(__file__))
import pytest
from OpenGL import acceleratesupport

class TestCoreDatatype(basetestcase.BaseTest):
    def test_arrayPointer( self ):
        dt = arraydatatype.GLuintArray
        d = dt.zeros( (3,))
        dp = dt.typedPointer( d )
        assert dp[0] == 0 
        assert dp[1] == 0
        assert dp[2] == 0
        dp[1] = 1
        assert dp[1] == 1
        assert d[1] == 1
    def test_ctypes_array( self ):
        color = (GLfloat * 3)( 0,1,0 )
        glColor3fv( color )
    if not OpenGL.ERROR_ON_COPY:
        def test_pointers( self ):
            """Test that basic pointer functions work"""
            vertex = GLdouble * 3
            vArray =  vertex * 2
            glVertexPointerd( [[2,3,4,5],[2,3,4,5]] )
            glVertexPointeri( ([2,3,4,5],[2,3,4,5]) )
            glVertexPointers( [[2,3,4,5],[2,3,4,5]] )
            glVertexPointerd( vArray( vertex(2,3,4),vertex(2,3,4) ) )
            myVector = vArray( vertex(2,3,4),vertex(2,3,4) )
            glVertexPointer(
                3,
                GL_DOUBLE,
                0,
                ctypes.cast( myVector, ctypes.POINTER(GLdouble)) 
            )
            
            repr(glVertexPointerb( [[2,3],[4,5]] ))
            glVertexPointerf( [[2,3],[4,5]] )
            assert arrays.ArrayDatatype.dataPointer( None ) == None
            glVertexPointerf( None )
            
            glNormalPointerd( [[2,3,4],[2,3,4]] )
            glNormalPointerd( None )
        
            glTexCoordPointerd( [[2,3,4],[2,3,4]] )
            glTexCoordPointerd( None )
        
            glColorPointerd( [[2,3,4],[2,3,4]] )
            glColorPointerd( None )
        
            glEdgeFlagPointerb( [0,1,0,0,1,0] )
            glEdgeFlagPointerb( None )
        
            glIndexPointerd( [0,1,0,0,1,0] )
            glIndexPointerd( None )
            
            glColor4fv( [0,0,0,1] )
            
            # string data-types...
            import struct
            s = struct.pack( '>iiii', 2,3,4,5 ) * 2
            glVertexPointer( 4,GL_INT,0,s )
        def test_texture( self ):
            """Test texture (requires OpenGLContext and PIL)"""
            try:
                from OpenGLContext import texture
                import Image 
                from OpenGL.GLUT import glutSolidTeapot
            except ImportError:
                pass
            else:
                assert glutSolidTeapot
                glEnable( GL_TEXTURE_2D )
                ourTexture = texture.Texture(
                    Image.open( os.path.join( HERE, 'yingyang.png') )
                )
                ourTexture()
                
                result = glGetTexImageub( GL_TEXTURE_2D,0,GL_RGBA )
                assert isinstance( result, bytes ), type(result)
                result = glGetTexImage( GL_TEXTURE_2D,0,GL_RGBA, GL_UNSIGNED_BYTE )
                assert isinstance( result, bytes ), type(result)
                
                glEnable( GL_LIGHTING )
                glEnable( GL_LIGHT0 )
                glBegin( GL_TRIANGLES )
                try:
                    try:
                        glTexCoord2f( .5, 1 )
                        glVertex3f( 0.,1.,0. )
                    except Exception:
                        traceback.print_exc()
                    glTexCoord2f( 0, 0 )
                    glVertex3fv( [-1,0,0] )
                    glTexCoord2f( 1, 0 )
                    glVertex3dv( [1,0,0] )
                    try:
                        glVertex3dv( [1,0] )
                    except ValueError:
                        assert OpenGL.ARRAY_SIZE_CHECKING, """Should have raised ValueError when doing array size checking"""
                    else:
                        assert not OpenGL.ARRAY_SIZE_CHECKING, """Should not have raised ValueError when not doing array size checking"""
                finally:
                    glEnd()
    @pytest.mark.skipif( not np, reason="Numpy not available")
    def test_numpyConversion( self ):
        """Test that we can run a numpy conversion from double to float for glColorArray"""
        a = np.arange( 0,1.2, .1, 'd' ).reshape( (-1,3 ))
        glEnableClientState(GL_VERTEX_ARRAY)
        try:
            glColorPointerf( a )
            glColorPointerd( a )
        finally:
            glDisableClientState( GL_VERTEX_ARRAY )
    @pytest.mark.skipif( not np, reason="Numpy not available")
    def test_glbuffersubdata_numeric(self):
        from OpenGL.arrays import vbo
        assert vbo.get_implementation()
        points = np.array( [
            [0,0,0],
            [0,1,0],
            [1,.5,0],
            [1,0,0],
            [1.5,.5,0],
            [1.5,0,0],
        ], dtype='f')
        d = vbo.VBO(points)
        with d:
            glBufferSubData(
                d.target,
                12,
                12,
                np.array([1,1,1],dtype='f'),
            )
    @pytest.mark.skipif( not np, reason="Numpy not available")
    @pytest.mark.skipif( OpenGL.ERROR_ON_COPY, reason="Test requires array copy")
    def test_copyNonContiguous( self ):
        """Test that a non-contiguous (transposed) array gets applied as a copy"""
        glMatrixMode(GL_MODELVIEW)
        glPushMatrix( )
        try:
            transf = np.identity(4, dtype=np.float32)
            # some arbitrary transformation...
            transf[0,3] = 2.5
            transf[2,3] = -80
            
            # what do we get with the un-transposed version...
            glMatrixMode(GL_MODELVIEW)
            glLoadIdentity()
            glMultMatrixf(transf)
            untransposed = glGetFloatv(GL_MODELVIEW_MATRIX)
            # now transposed...

            # with a copy it works...
            t2 = transf.transpose().copy()
            # This doesn't work:
            glLoadIdentity()
            glMultMatrixf(t2)
            # This does work:
            #glMultMatrixf(transf.transpose().copy())
            transposed = glGetFloatv(GL_MODELVIEW_MATRIX)

            assert not np.allclose( transposed, untransposed ), (transposed, untransposed)
            
            t2 = transf.transpose()
            # This doesn't work:
            glLoadIdentity()
            glMultMatrixf(t2)
            # This does work:
            #glMultMatrixf(transf.transpose().copy())
            transposed = glGetFloatv(GL_MODELVIEW_MATRIX)
            
            assert not np.allclose( transposed, untransposed ), (transposed, untransposed)
        finally:
            glMatrixMode(GL_MODELVIEW)
            glPopMatrix()

    def test_bytes_array_support( self ):
        color = b'\000'*12
        glColor3fv( color )
        
        
    @pytest.mark.skipif( not (
        OpenGL.USE_ACCELERATE and acceleratesupport.ACCELERATE_AVAILABLE
    ), reason="Need OpenGL_accelerate for buffer support")
    def test_bytearray_support( self ):
        import struct 
        data = struct.pack( b'fff', .5, .4, .3 )
        color = bytearray( data )
        glColor3fv( color )
    
    @pytest.mark.skipif( not (
        OpenGL.USE_ACCELERATE and acceleratesupport.ACCELERATE_AVAILABLE
    ), reason="Need OpenGL_accelerate for buffer support")
    def test_buffer_api_basic(self):
        import array as silly_array
        structures = [
            (b'this and that',13,1,True,1,b'B',[13],[1]),
        ]
        if sys.version_info[:2] >= (2,7):
            structures.append(
                # on Python 3.4 we do *not* get the (3) prefix :(
                ((GLint * 3)( 1,2,3 ),12,4,False,1,[b'(3)<i',b'(3)<l',b'<i',b'<l'],[3],None),
            )
        
        if sys.version_info[:2] >= (3,0):
            # only supports buffer protocol in 3.x
            structures.extend([
                (silly_array.array('I',[1,2,3]),12,4,False,1,b'I',[3],[4]),
            ])
        try:
            structures.append( (memoryview(b'this'),4,1,True,1,b'B',[4],[1]) )
        except NameError:
            # Python 2.6 doesn't have memory view 
            pass
        try:
            if array:
                structures.extend( [
                    (arange(0,9,dtype='I').reshape((3,3)),36,4,False,2,b'I',[3,3],[12,4]),
                    (arange(0,9,dtype='I').reshape((3,3))[:,1],12,4,False,1,b'I',[3],[12]),
                ])
        except NameError:
            # Don't have numpy installed...
            pass
        
        from OpenGL.arrays import _buffers
        for object,length,itemsize,readonly,ndim,format,shape,strides in structures:
            buf = _buffers.Py_buffer.from_object( object, _buffers.PyBUF_STRIDES|_buffers.PyBUF_FORMAT )
            with buf:
                assert buf.len == length, (object,length,buf.len)
                assert buf.itemsize == itemsize, (object,itemsize,buf.itemsize)
                assert buf.readonly == readonly, (object,readonly,buf.readonly)
                assert buf.ndim == ndim, (object,ndim,buf.ndim)
                if isinstance( format, list):
                    assert buf.format in format, (object,format,buf.format)
                else:
                    assert buf.format == format, (object,format,buf.format)
                assert buf.shape[:buf.ndim] == shape, (object, shape, buf.shape[:buf.ndim])
                assert buf.dims == shape, (object, shape, buf.dims )
                assert buf.buf 
                if strides is None:
                    assert not buf.strides 
                else:
                    assert buf.strides[:buf.ndim] == strides, (object, strides, buf.strides[:buf.ndim])
            assert buf.obj == None, buf.obj
            del buf
    
    
    @pytest.mark.skipif( not (
        OpenGL.USE_ACCELERATE and acceleratesupport.ACCELERATE_AVAILABLE
    ), reason="Need OpenGL_accelerate for buffer support")
    def test_memoryview_support( self ):
        color = bytearray( b'\000'*12 )
        mem = memoryview( color )
        glColor3fv( mem )
                

    def test_void_dp_for_void_dp_is_self( self ):
        array = ctypes.c_voidp( 12 )
        translated = ArrayDatatype.voidDataPointer( array )
        assert translated.value == array.value, translated

    def test_params_python3_strings( self ):
        try:
            glGetUniformBlockIndex( 0, unicode("Moo") )
        except ArgumentError:
            assert OpenGL.ERROR_ON_COPY, """Shouldn't have raised error on copy for unicode"""
        except TypeError:
            raise
        except GLError:
            # expected error, as we don't have a shader there...
            pass

    @pytest.mark.skipif( not np, reason="Numpy not available")
    def test_array_subclass( self ):
        s = Subclassed([0,1,2,3,4])
        result = arraydatatype.ArrayDatatype.asArray( s )
        assert isinstance( result, Subclassed )
    
    @pytest.mark.skipif( not np, reason="Numpy not available")
    def test_byte_count_numpy( self ):
        for a,expected in [
            (np.array([1,2],dtype='e'),4),
            (np.array([1,2],dtype='f'),8),
            (np.array([1,2],dtype='d'),16),
            (np.array([1,2],dtype='B'),2),
            (np.array([[1],[2]],dtype='B'),2),
            (np.float32(),4),
            (np.float64(),8),
        ]:
            handler = arraydatatype.ArrayDatatype.getHandler( a )
            
            assert type(a) in handler.HANDLED_TYPES, type(a)
            calculated = arraydatatype.ArrayDatatype.arrayByteCount( a )
            assert calculated == expected, "Byte count for %s was %s, expected %"%(
                a, calculated, expected,
            )
    def test_byte_count( self ):
        for a,expected in [
            (ctypes.c_float(),4),
            ((ctypes.c_float*3*4)(),4*3*4),
            
        ]:
            calculated = arraydatatype.ArrayDatatype.arrayByteCount( a )
            assert calculated == expected, "Byte count for %s was %s, expected %"%(
                a, calculated, expected,
            )
        
if np:
    class Subclassed( np.ndarray ):
        pass 



from __future__ import print_function
import os
if not os.environ.get( 'PYOPENGL_PLATFORM' ):
    os.environ['PYOPENGL_PLATFORM'] = 'osmesa'
import OpenGL
OpenGL.USE_ACCELERATE = False
from OpenGL.GL import *
from OpenGL.arrays import vbo
from OpenGL import extensions
from OpenGL.GL.ARB.fragment_program import glGenProgramsARB
from OpenGL.GLU import *
from OpenGL.osmesa import *
from OpenGL.GL import shaders
from numpy import arange, array
import contextlib

def write_ppm(buf, filename):
    """Write traditional ASCII PPM file from buffer"""
    with open(filename, "w") as f:
        h, w, c = buf.shape
        print( "P3", file=f)
        print( "# ascii ppm file created by osmesa",file=f)
        print( "%i %i" % (w, h),file=f)
        print("255",file=f)
        for y in range(h - 1, -1, -1):
            for x in range(w):
                pixel = buf[y, x]
                l = " %3d %3d %3d" % (pixel[0], pixel[1], pixel[2])
                f.write(l)
            f.write("\n")

@contextlib.contextmanager
def osmesa_context(width=400, height=400, output='output.ppm'):
    """Do OSMesa rendering context setup and write to output if non-null"""
    from OpenGL import arrays
    ctx = OSMesaCreateContext(OSMESA_RGBA, None)
    buf = arrays.GLubyteArray.zeros((height, width, 4))
    assert(OSMesaMakeCurrent(ctx, buf, GL_UNSIGNED_BYTE, width, height))
    assert(OSMesaGetCurrentContext())
    yield (ctx, buf)
    if output:
        write_ppm(buf,output)
    OSMesaDestroyContext(ctx)

def setup():
    """Setup the rendering passes
    
    Taken from `OpenGLContext/tests/shader_4.py`
    """
    vertex = shaders.compileShader(
        """
        uniform float tween;
        attribute vec3 position;
        attribute vec3 tweened;
        attribute vec3 color;
        varying vec4 baseColor;
        void main() {
            gl_Position = gl_ModelViewProjectionMatrix * mix(
                vec4( position,1.0),
                vec4( tweened,1.0),
                tween
            );
            baseColor = vec4(color,1.0);
        }""",
        GL_VERTEX_SHADER
    )
    fragment = shaders.compileShader(
        """
        varying vec4 baseColor;
        void main() {
            gl_FragColor = baseColor;
        }""",
        GL_FRAGMENT_SHADER
    )
    shader = shaders.compileProgram(vertex,fragment)
    vertices = vbo.VBO(
        array( [
            [  0, 1, 0, 1,3,0,  0,1,0 ],
            [ -1,-1, 0, -1,-1,0,  1,1,0 ],
            [  1,-1, 0, 1,-1,0, 0,1,1 ],
            
            [  2,-1, 0, 2,-1,0, 1,0,0 ],
            [  4,-1, 0, 4,-1,0, 0,1,0 ],
            [  4, 1, 0, 4,9,0, 0,0,1 ],
            [  2,-1, 0, 2,-1,0, 1,0,0 ],
            [  4, 1, 0, 1,3,0, 0,0,1 ],
            [  2, 1, 0, 1,-1,0, 0,1,1 ],
        ],'f')
    )
    position_location = glGetAttribLocation( 
        shader, 'position' 
    )
    tweened_location = glGetAttribLocation(
        shader, 'tweened',
    )
    color_location = glGetAttribLocation( 
        shader, 'color' 
    )
    tween_location = glGetUniformLocation(
        shader, 'tween',
    )
    return {
        'shader': shader,
        'position_location': position_location,
        'tweened_location': tweened_location,
        'color_location': color_location,
        'tween_location': tween_location,
        'vertices': vertices,
    }
def render(names, fraction=.5):
    glUseProgram(names['shader'])
    glUniform1f(names['tween_location'],  fraction)
    with names['vertices']:
        try:
            glEnableVertexAttribArray( names['position_location'] )
            glEnableVertexAttribArray( names['tweened_location'] )
            glEnableVertexAttribArray( names['color_location'] )
            stride = 9*4
            glVertexAttribPointer( 
                names['position_location'], 
                3, GL_FLOAT,False, stride, names['vertices'] 
            )
            glVertexAttribPointer( 
                names['tweened_location'], 
                3, GL_FLOAT,False, stride, names['vertices']+12
            )
            glVertexAttribPointer( 
                names['color_location'], 
                3, GL_FLOAT,False, stride, names['vertices']+24
            )
            glDrawArrays(GL_TRIANGLES, 0, 9)
        finally:
            glDisableVertexAttribArray( names['position_location'] )
            glDisableVertexAttribArray( names['tweened_location'] )
            glDisableVertexAttribArray( names['color_location'] )

def main():
    with osmesa_context(400,400) as context:
        print('Queried extensions:', sorted(extensions.GLQuerier.pullExtensions()))
        print('Queried version:', extensions.GLQuerier.pullVersion())
        names = setup()
        for fraction in arange(0,.5,.1):
            glClearColor(0.0,0.0,1.0,1.0)
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT )
            print("Rendering: %s"%(fraction,))
            render(names,fraction)
            glFinish()


if __name__ == '__main__':
    main()


import OpenGL
OpenGL.ERROR_CHECKING = False
OpenGL.USE_ACCELERATE = False

from OpenGL.GLUT import *

def reshape(width, height): pass
def display(): glutSwapBuffers()

def main():
    glutInit([])
    glutInitDisplayMode(GLUT_RGBA|GLUT_3_2_CORE_PROFILE)
    glutCreateWindow(b"test")
    glutReshapeFunc(reshape)
    glutDisplayFunc(display)

    from OpenGL.GL import glGenVertexArrays, glVertex3f
    assert bool(glGenVertexArrays)
    try:
        if fgDeinitialize: fgDeinitialize(False)
    except NameError as err:
        pass # Older PyOpenGL, you may see a seg-fault here...

if __name__ == "__main__":
    main()



'''OpenGL extension NV.sample_mask_override_coverage

This module customises the behaviour of the 
OpenGL.raw.GLES2.NV.sample_mask_override_coverage to provide a more 
Python-friendly API

Overview (from the spec)
	
	This extension allows the fragment shader to control whether the
	gl_SampleMask output can enable samples that were not covered by the
	original primitive, or that failed the early depth/stencil tests.
	This can be enabled by redeclaring the gl_SampleMask output with the
	"override_coverage" layout qualifier:
	
	    layout(override_coverage) out int gl_SampleMask[];

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/NV/sample_mask_override_coverage.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.GLES2 import _types, _glgets
from OpenGL.raw.GLES2.NV.sample_mask_override_coverage import *
from OpenGL.raw.GLES2.NV.sample_mask_override_coverage import _EXTENSION_NAME

def glInitSampleMaskOverrideCoverageNV():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )


### END AUTOGENERATED SECTION


from __future__ import print_function
from OpenGL.GL import *
from OpenGL.GLUT import *
from OpenGL.extensions import alternate
from OpenGL.GL.ARB.draw_instanced import *
from OpenGL.GL.EXT.draw_instanced import *

def detect():
    glutInit([])
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH)
    glutInitWindowSize(200,200)

    glutCreateWindow("Detection")
    
    functions = (
        glDrawArraysInstanced, 
        glDrawArraysInstancedARB, 
        glDrawArraysInstancedEXT,
        #glDrawArraysInstancedNV, is GLES only
    )
    
    for function in functions:
        print(function.__name__, bool(function))
    
    any = alternate( *functions )
    
    print(any, bool(any))
    try:
        if fgDeinitialize: fgDeinitialize(False)
    except NameError:
        pass # Older PyOpenGL, you may see a seg-fault here...
    

if __name__ == "__main__":
    detect()



#! /usr/bin/env python
import egltest
import numpy
from OpenGL import GL

@egltest.egltest( api='opengl' )
def test_gl( ):
    GL.glClearColor( 1,0,0, 0 )
    GL.glClear( GL.GL_COLOR_BUFFER_BIT|GL.GL_DEPTH_BUFFER_BIT )

if __name__ == "__main__":
    test_gl()
    



#! /usr/bin/env python
"""Generate python PyOpenGL api from xml registry documents (using xmlreg)"""
from __future__ import print_function
import os, logging, subprocess, glob
import xmlreg, codegenerator
import ctypetopytype
log = logging.getLogger( 'xml-generate' )

KHRONOS_URL = 'https://github.com/KhronosGroup/OpenGL-Registry.git'
KHRONOS_API = os.path.join( os.path.dirname(__file__), 'khronosapi' )

def get_khronos( khronosapi ):
    if not os.path.exists( khronosapi ):
        subprocess.check_call(['git','clone', KHRONOS_URL, khronosapi ])

def main(khronosapi=None):
    khronosapi = khronosapi or KHRONOS_API
    get_khronos( khronosapi )
    
    files = sorted( glob.glob( os.path.join( khronosapi, 'xml','*.xml' ) ))
    for file in files:
        generate_for_file( file )

def generate_for_file( filename ):
    log.info( 'Starting file: %s', filename )
    registry = xmlreg.parse( filename )
    generator = codegenerator.Generator(
        registry,
        ctypetopytype.ctype_to_pytype
    )
    for name,feature in registry.feature_set.items():
        print(feature.name, feature.api)
        generator.module( feature )
    for name,extension in registry.extension_set.items():
        print(extension.name, extension.apis)
        generator.module( extension )

if __name__ == "__main__":
    logging.basicConfig( level=logging.INFO )
    main()
    



"""Transliteration of https://open.gl/feedback into Python"""
from __future__ import print_function
import pygamegltest
from OpenGL._bytes import as_8_bit
from OpenGL.GL import *
from OpenGL.GL import shaders
vertex_shader = """#version 150 core
    in float inValue;
    out float geoValue;
    uniform float junk; // Just here for the introspection test

    void main() {
        geoValue = sqrt(inValue) + junk;
    }
"""

geometry_shader = """#version 150 core
    layout(points) in;
    layout(triangle_strip, max_vertices = 3) out;

    in float[] geoValue;
    out float outValue;

    void main() {
        for (int i = 0; i < 3; i++) {
            outValue = geoValue[0] + i;
            EmitVertex();
        }

        EndPrimitive();
    }
"""

@pygamegltest.pygametest(name="Geometry Shader Test")
def main():
    program = shaders.compileProgram(
        shaders.compileShader([vertex_shader], GL_VERTEX_SHADER), 
        shaders.compileShader([geometry_shader], GL_GEOMETRY_SHADER)
    )
    
    # test that we can describe the attributes in the shader
    for i in range( glGetProgramiv( program, GL_ACTIVE_ATTRIBUTES )):
        name,size,type = glGetActiveAttrib( program, i )
        print( 'Arribute %s(%i) -> %s %s'%(name,i,size,type))
        
        length,size,type,name = GLsizei(),GLint(),GLenum(),(GLchar*8)()
        glGetActiveAttrib( program, i, None, length,size,type,name )
        assert length.value == len(name.value),(length.value,name.value)
        assert type.value == 5126, type.value
        
    for i in range( glGetProgramiv( program, GL_ACTIVE_UNIFORMS )):
        name,size,type = glGetActiveUniform( program, i )
        print( 'Uniform %s(%i) -> %s %s'%(name,i,size,type))
        
        length,size,type,name = GLsizei(),GLint(),GLenum(),(GLchar*5)()
        glGetActiveUniform( program, i, 5, length, size, type, name )
        assert length.value == len(name.value),(length.value,name.value)
        assert type.value == 5126, type.value
        

    buff = (ctypes.c_char_p * 1)( as_8_bit("outValue\000") )
    c_text = ctypes.cast(ctypes.pointer(buff), ctypes.POINTER(ctypes.POINTER(GLchar)))
    # modifies the state in the linking of the program
    glTransformFeedbackVaryings(program, 1, c_text, GL_INTERLEAVED_ATTRIBS);
    # so have to re-link
    glLinkProgram(program)
    

    glUseProgram(program);

    vao = glGenVertexArrays(1);
    glBindVertexArray(vao);

    data = (GLfloat * 5)(1.0, 2.0, 3.0, 4.0, 5.0)
    
    vbo = glGenBuffers(1);
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    glBufferData(GL_ARRAY_BUFFER, ctypes.sizeof(data), data, GL_STATIC_DRAW);

    inputAttrib = glGetAttribLocation(program, "inValue");
    glEnableVertexAttribArray(inputAttrib);
    glVertexAttribPointer(inputAttrib, 1, GL_FLOAT, GL_FALSE, 0, 0);

    tbo = glGenBuffers(1);
    glBindBuffer(GL_ARRAY_BUFFER, tbo);
    glBufferData(GL_ARRAY_BUFFER, ctypes.sizeof(data) * 3, None, GL_STATIC_READ);

    glEnable(GL_RASTERIZER_DISCARD);

    glBindBufferBase(GL_TRANSFORM_FEEDBACK_BUFFER, 0, tbo);

    glBeginTransformFeedback(GL_TRIANGLES);
    glDrawArrays(GL_POINTS, 0, 5);
    glEndTransformFeedback();

    glDisable(GL_RASTERIZER_DISCARD);

    glFlush();

    feedback = (GLfloat*15)(*([0]*15))
    glGetBufferSubData(GL_TRANSFORM_FEEDBACK_BUFFER, 0, ctypes.sizeof(feedback), feedback);

    for value in feedback:
        print(value)

if __name__ == "__main__":
    main()



from __future__ import print_function
import ctypes
try:
    from ctypes.util import find_library
except ImportError as err:
    from ctypes.util import findLib 
    def find_library( string ):
        return findLib( string )[0]

GL = OpenGL = ctypes.CDLL( find_library('GL'), mode=ctypes.RTLD_GLOBAL )
GLU = ctypes.CDLL( find_library('GLU'), mode=ctypes.RTLD_GLOBAL )
# glut shouldn't need to be global IIUC
GLUT = ctypes.CDLL( find_library('glut'), mode=ctypes.RTLD_GLOBAL )
print('GL', GL)
print('GLU', GLU)
print('GLUT', GLUT)
print(GLUT.glutSolidTeapot)



from __future__ import print_function
#import OpenGL
#OpenGL.USE_ACCELERATE=False
from OpenGL.GL import *
from OpenGL.GLX import *
from OpenGL.GLX.EXT.texture_from_pixmap import *
from pygamegltest import pygametest
import os

# print('Not yet working')
# raise SystemExit(1)
attributes = [
#    GLX_BIND_TO_TEXTURE_RGBA_EXT, 1,
#    GLX_DRAWABLE_TYPE, GLX_PIXMAP_BIT,
#    GLX_BIND_TO_TEXTURE_TARGETS_EXT, GLX_TEXTURE_2D_BIT_EXT,
    GLX_DRAWABLE_TYPE, GLX_WINDOW_BIT,
    GLX_RED_SIZE, 8,
    GLX_GREEN_SIZE, 8,
    GLX_BLUE_SIZE, 8,
    GLX_ALPHA_SIZE, 8,
    GLX_DEPTH_SIZE, 24,
    GLX_STENCIL_SIZE, 8,
    GLX_RENDER_TYPE, GLX_RGBA_BIT,
    GLX_X_VISUAL_TYPE , GLX_TRUE_COLOR,
    GLX_DOUBLEBUFFER, 1,
    GLX_Y_INVERTED_EXT, GLX_DONT_CARE,
    GL_NONE
]
attributes = (GLint * len(attributes))( * attributes )

import ctypes
from OpenGL.platform import ctypesloader

X11 = ctypesloader.loadLibrary( ctypes.cdll, 'X11' )

XDefaultScreen = X11.XDefaultScreen
XDefaultScreen.argtypes = [ctypes.POINTER(Display)]

XOpenDisplay = X11.XOpenDisplay 
XOpenDisplay.restype = ctypes.POINTER(Display)

XRootWindow = X11.XRootWindow
XRootWindow.restyle = ctypes.POINTER( Window )

XCreateWindow = X11.XCreateWindow
XCreateWindow.restyle = ctypes.POINTER( Window )
XCreateWindow.argtypes = [
    ctypes.POINTER(Display),ctypes.POINTER(Window),
    GLint,GLint,GLuint,GLuint,GLuint,GLint,GLuint,
    ctypes.POINTER(Visual),
    ctypes.c_ulong,ctypes.c_void_p,
]
XFree = X11.XFree
AllocNone = 0
def debug_struct(s):
    return dict([
        (k, getattr(s,k,None))
        for k in s.__class__.__slots__
    ])

#@pygametest()
def main():
    display = XOpenDisplay( os.environ.get( 'DISPLAY' ))
    if not display:
        raise RuntimeError("Unable to get the default display")
    screen = XDefaultScreen( display )
    print('X Display %s Screen %s'%( display, screen ))
    major,minor = GLint(),GLint()
    glXQueryVersion(display, major, minor)
    version = (major.value,minor.value)
    print('glX Version: %s.%s'%version)
    if (major.value,minor.value) < (1,3):
        print("Need at least GLX 1.3 to choose the framebuffer config")
        raise RuntimeError((major.value,minor.value))

    # Get the framebuffer configuration...
    count = ctypes.c_int(0)
    configs = glXChooseFBConfig(display, screen, attributes, ctypes.pointer(count))
    if count.value < 1:
        raise RuntimeError('Did not find any configs')
    print('Found %s configs'%(count.value,))

    for index in range(count.value):
        vis = glXGetVisualFromFBConfig( display, configs[index] )
        if vis:
            vis = vis[0]
            print('Visual %s: %s'%(index+1, debug_struct(vis)))
    
    # get a visual with 1.0 functionality...
    if not vis:
        print("Did not get a double-buffering visual, somehow?")
        raise RuntimeError("No double-buffered visual available")
    
    root = XRootWindow(display,vis.screen)
    root_p = ctypes.c_ulong(root)
    window = XCreateWindow( 
        display, 
        root_p, 
        0,0, #x,y
        300,300, #w,h,
        1, # border width
        vis.depth,
        1, # InputOutput Class
        vis.visual,
        0,
        ctypes.c_void_p(0),
    )
        
    context = glXCreateContext(display,vis,None,GL_TRUE)
    
    if version >= (1,1):
        print(glXQueryExtensionsString(display,screen))
#        if version >= (1,2):
#            d = glXGetCurrentDisplay()[0]
#            print 'Current display', d
#        else:
    
    
    
    if version >= (1,3):
        elements = GLint(0)
        configs = glXChooseFBConfig(
            display, 
            screen, 
            attributes, 
            elements
        )
        print('%s configs found'%( elements.value ))
        for config in range( elements.value ):
            print('Config: %s %s'%(config,configs[config][0]))
            samples = ctypes.c_int()
            for attribute in (
                'GLX_FBCONFIG_ID','GLX_BUFFER_SIZE',
                'GLX_LEVEL','GLX_DOUBLEBUFFER',
                'GLX_STEREO',
                'GLX_SAMPLES','GLX_SAMPLE_BUFFERS',
                'GLX_DRAWABLE_TYPE',
            ):
                glXGetFBConfigAttrib( display, configs[config], globals()[attribute], samples )
                print('%s -> %s'%( attribute, samples.value ))
            print() 
    
        
if __name__ == "__main__":
    main()



#! /usr/bin/env python
from __future__ import print_function
import os
if not os.environ.get( 'PYOPENGL_PLATFORM' ):
    os.environ['PYOPENGL_PLATFORM'] = 'osmesa'
from math import pi, sin, cos
import OpenGL
OpenGL.USE_ACCELERATE = False
from OpenGL.GL import *
from OpenGL.GL.ARB.fragment_program import glGenProgramsARB
from OpenGL.GLU import *
from OpenGL.osmesa import *

def sphere(radius, slices, stacks):
    q = gluNewQuadric()
    gluQuadricNormals(q, GLU_SMOOTH)
    gluSphere(q, radius, slices, stacks)
    gluDeleteQuadric(q)


def cone(base, height, slices, stacks):
    q = gluNewQuadric()
    gluQuadricDrawStyle(q, GLU_FILL)
    gluQuadricNormals(q, GLU_SMOOTH)
    gluCylinder(q, base, 0.0, height, slices, stacks)
    gluDeleteQuadric(q)


def torus(innerRadius, outerRadius, sides, rings):
    ringDelta = 2.0 * pi / rings
    sideDelta = 2.0 * pi / sides

    theta = 0.0
    cosTheta = 1.0
    sinTheta = 0.0
    for i in range(rings - 1, -1, -1):
        theta1 = theta + ringDelta
        cosTheta1 = cos(theta1)
        sinTheta1 = sin(theta1)
        glBegin(GL_QUAD_STRIP)
        phi = 0.0
        for j in range(sides, -1, -1):
            phi += sideDelta
            cosPhi = cos(phi)
            sinPhi = sin(phi)
            dist = outerRadius + innerRadius * cosPhi

            glNormal3f(cosTheta1 * cosPhi, -sinTheta1 * cosPhi, sinPhi)
            glVertex3f(cosTheta1 * dist, -sinTheta1 * dist, innerRadius * sinPhi)
            glNormal3f(cosTheta * cosPhi, -sinTheta * cosPhi, sinPhi)
            glVertex3f(cosTheta * dist, -sinTheta * dist,  innerRadius * sinPhi)
        glEnd()
        theta = theta1
        cosTheta = cosTheta1
        sinTheta = sinTheta1


def render_image():
    light_ambient = [0.0, 0.0, 0.0, 1.0]
    light_diffuse = [1.0, 1.0, 1.0, 1.0]
    light_specular = [1.0, 1.0, 1.0, 1.0]
    light_position = [1.0, 1.0, 1.0, 0.0]
    red_mat = [1.0, 0.2, 0.2, 1.0]
    green_mat = [0.2, 1.0, 0.2, 1.0]
    blue_mat = [0.2, 0.2, 1.0, 1.0]

    glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient)
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse)
    glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular)
    glLightfv(GL_LIGHT0, GL_POSITION, light_position)

    glEnable(GL_LIGHTING)
    glEnable(GL_LIGHT0)
    glEnable(GL_DEPTH_TEST)

    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()
    glOrtho(-2.5, 2.5, -2.5, 2.5, -10.0, 10.0)
    glMatrixMode(GL_MODELVIEW)

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

    glPushMatrix()
    glRotatef(20.0, 1.0, 0.0, 0.0)

    glPushMatrix()
    glTranslatef(-0.75, 0.5, 0.0)
    glRotatef(90.0, 1.0, 0.0, 0.0)
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, red_mat)
    torus(0.275, 0.85, 20, 20)
    glPopMatrix()

    glPushMatrix()
    glTranslatef(-0.75, -0.5, 0.0)
    glRotatef(270.0, 1.0, 0.0, 0.0)
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, green_mat)
    cone(1.0, 2.0, 16, 1)
    glPopMatrix()

    glPushMatrix()
    glTranslatef(0.75, 0.0, -1.0)
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, blue_mat)
    sphere(1.0, 20, 20)
    glPopMatrix()

    glPopMatrix()

    # This is very important!!!
    # Make sure buffered commands are finished!!!
    glFinish()


def write_ppm(buf, filename):
    f = open(filename, "w")
    if f:
        h, w, c = buf.shape
        print( "P3", file=f)
        print( "# ascii ppm file created by osmesa",file=f)
        print( "%i %i" % (w, h),file=f)
        print("255",file=f)
        for y in range(h - 1, -1, -1):
            for x in range(w):
                pixel = buf[y, x]
                l = " %3d %3d %3d" % (pixel[0], pixel[1], pixel[2])
                f.write(l)
            f.write("\n")

if __name__ == '__main__':
    from OpenGL import arrays

    ctx = OSMesaCreateContext(OSMESA_RGBA, None)
    #ctx = OSMesaCreateContextExt(OSMESA_RGBA, 32, 0, 0, None)

    width, height = 400, 400
    buf = arrays.GLubyteArray.zeros((height, width, 4))
    assert(OSMesaMakeCurrent(ctx, buf, GL_UNSIGNED_BYTE, width, height))

    assert(OSMesaGetCurrentContext())
    
    program = glGenProgramsARB(1)
    assert(program)

    z = glGetIntegerv(GL_DEPTH_BITS)
    s = glGetIntegerv(GL_STENCIL_BITS)
    a = glGetIntegerv(GL_ACCUM_RED_BITS)
    print("OpenGL Version: %s"%(extensions.GLQuerier.getVersion()))
    print("Depth=%d Stencil=%d Accum=%d" % (z, s, a))

    print("Width=%d Height=%d" % (OSMesaGetIntegerv(OSMESA_WIDTH),
                                  OSMesaGetIntegerv(OSMESA_HEIGHT)))
    #OSMesaPixelStore(OSMESA_Y_UP, 0)

    render_image()

    write_ppm(buf, 'output.ppm')

    OSMesaDestroyContext(ctx)



import basetestcase
import os
from OpenGL.GL import *
from OpenGL.GLU import *
try:
    import numpy as np
except ImportError as err:
    np = None
HERE = os.path.abspath(os.path.dirname(__file__))

class TestTess( basetestcase.BaseTest ):
    TESS_TEST_SHAPE = [
            [191,   0],
            [ 191, 1480],
            [ 191, 1480],
            [ 401, 1480],
            [ 401, 1480],
            [401,   856],
            [401,   856],
            [1105,  856],
            [1105,  856],
            [1105, 1480],
            [1105, 1480],
            [1315, 1480],
            [1315, 1480],
            [1315,    0],
            [1315,    0],
            [1105,    0],
            [1105,    0],
            [1105,  699],
            [1105,  699],
            [401,   699],
            [401,   699],
            [401,     0],
            [401,     0],
            [191,     0],
            [191,     0],
            [191,     0],
        ]
    def test_tess(self ):
        """Test that tessellation works"""
        glDisable( GL_LIGHTING )
        glColor3f( 1,1,1 )
        glNormal3f( 0,0,1 )
        def begin( *args ):
            return glBegin( *args )
        def vertex( *args ):
            return glVertex3dv( *args )
        def end( *args ):
            return glEnd( *args )
        def combine( coords, vertex_data, weight):
            return coords
        tobj = gluNewTess()
        gluTessCallback(tobj, GLU_TESS_BEGIN, begin);
        gluTessCallback(tobj, GLU_TESS_VERTEX, vertex); 
        gluTessCallback(tobj, GLU_TESS_END, end); 
        gluTessCallback(tobj, GLU_TESS_COMBINE, combine); 
        gluTessBeginPolygon(tobj, None); 
        gluTessBeginContour(tobj);
        for (x,y) in self.TESS_TEST_SHAPE:
            vert = (x,y,0.0)
            gluTessVertex(tobj, vert, vert);
        gluTessEndContour(tobj); 
        gluTessEndPolygon(tobj);
    def test_tess_collection( self ):
        """SF#2354596 tessellation combine results collected"""
        all_vertices = []
        combinations = []
        def start(*args):
            pass
        def stop(*args):
            pass
        def tessvertex(vertex_data, polygon_data=None):
            # polygon data *should* be collected here
            #assert polygon_data is all_vertices, polygon_data
            all_vertices.append(vertex_data)
            #collected.append( vertex_data )
            return polygon_data
        def tesscombine(coords, vertex_data, weights,_=None):
            new = (True,coords)
            combinations.append( coords )
            return new

        def tessedge(flag,*args,**named):
            pass	# dummy
        def tesserr( enum ):
            raise RuntimeError( gluErrorString( enum ) )

        # set up tessellator in CSG intersection mode
        tess=gluNewTess()
        gluTessProperty(tess, GLU_TESS_WINDING_RULE, GLU_TESS_WINDING_ABS_GEQ_TWO)
        gluTessCallback(tess, GLU_TESS_BEGIN, start)
        gluTessCallback(tess, GLU_TESS_END, stop)
        gluTessCallback(tess, GLU_TESS_EDGE_FLAG, tessedge)	# no strips
        gluTessCallback(tess, GLU_TESS_VERTEX, tessvertex)
        gluTessCallback(tess, GLU_TESS_ERROR, tesserr )
        gluTessCallback(tess, GLU_TESS_COMBINE, tesscombine)

        gluTessBeginPolygon(tess, all_vertices)
        try:
            for contour in [
                # first square
                [(-1,0,-1),(1,0,-1),(1,0,1),(-1,0,1)],
                # second intersects the first
                [(.5,0,-.5),(1.5,0,-.5),(1.5,0,.5),(.5,0,.5)],
            ]:
                
                gluTessBeginContour(tess)
                try:
                    for point in contour:
                        if np:
                            if OpenGL.ERROR_ON_COPY:
                                point = np.array( point, 'd' )
                            else:
                                point = np.array( point, 'f' )
                        else: 
                            point = (GLdouble*3)(*point)
                        gluTessVertex( tess, point, (False,point))
                finally:
                    gluTessEndContour(tess)
        finally:
            gluTessEndPolygon(tess)

        # Show collected triangle vertices :-
        # Original input vertices are marked as False.
        # Vertices generated in combine callback are marked as True.
        assert all_vertices, "Nothing collected"
        combined,original = [x for x in all_vertices if x[0]], [x for x in all_vertices if not x[0]]
        
        assert combined, ("No combined vertices", all_vertices )
        assert original, ("No original vertices", all_vertices )
        assert len(combinations) == 2, combinations
    
    def test_tess_cb_traditional( self ):
        outline = [
            [191,   0],
            [ 191, 1480],
            [ 191, 1480],
            [ 401, 1480],
            [ 401, 1480],
            [401,   856],
            [401,   856],
            [1105,  856],
            [1105,  856],
            [1105, 1480],
            [1105, 1480],
            [1315, 1480],
            [1315, 1480],
            [1315,    0],
            [1315,    0],
            [1105,    0],
            [1105,    0],
            [1105,  699],
            [1105,  699],
            [401,   699],
            [401,   699],
            [401,     0],
            [401,     0],
            [191,     0],
            [191,     0],
            [191,     0],
        ]
        scale = 1200.
        self.tess = gluNewTess()
        gluTessCallback(self.tess, GLU_TESS_BEGIN, glBegin)
        def test( t, polyData=None ):
            glNormal( 0,0, -1 )
            glColor3f( t[0],t[1],t[2] )
            return glVertex3f( t[0],t[1],t[2])
        gluTessCallback(self.tess, GLU_TESS_VERTEX_DATA, test)
        gluTessCallback(self.tess, GLU_TESS_END, glEnd);
        combined = []
        def combine( points, vertices, weights ):
            #print 'combine called', points, vertices, weights
            combined.append( points )
            return points
        gluTessCallback(self.tess, GLU_TESS_COMBINE, combine)
        gluTessBeginPolygon( self.tess, None )
        try:
            gluTessBeginContour( self.tess )
            try:
                for (x,y) in outline:
                    if np:
                        vertex = np.array((x/scale,y/scale,0.0),'d')
                    else:
                        vertex = (GLdouble*3)(x/scale,y/scale,0.0)
                    gluTessVertex(self.tess, vertex, vertex)
            finally:
                gluTessEndContour( self.tess )
        finally:
            gluTessEndPolygon(self.tess)



"""EGL Pygame test framework"""
from __future__ import print_function
import os
if not os.environ.get( 'PYOPENGL_PLATFORM' ):
    os.environ['PYOPENGL_PLATFORM'] = 'egl'
import ctypes
import pygame.display 
import pygame 
import logging
import OpenGL
from functools import wraps

if os.environ.get( 'TEST_NO_ACCELERATE' ):
    OpenGL.USE_ACCELERATE = False

from OpenGL import arrays
from OpenGL.EGL import *
log = logging.getLogger( __name__ )

DESIRED_ATTRIBUTES = [
    EGL_BLUE_SIZE, 8,
    EGL_RED_SIZE,8,
    EGL_GREEN_SIZE,8,
    EGL_DEPTH_SIZE,24,
    EGL_COLOR_BUFFER_TYPE, EGL_RGB_BUFFER,
    EGL_CONFIG_CAVEAT, EGL_NONE, # Don't allow slow/non-conformant
]
API_BITS = {
    'opengl': EGL_OPENGL_BIT,
    'gl': EGL_OPENGL_BIT,
    'gles2': EGL_OPENGL_ES2_BIT,
    'gles1': EGL_OPENGL_ES_BIT,
    'gles': EGL_OPENGL_ES_BIT,
    'es2': EGL_OPENGL_ES2_BIT,
    'es1': EGL_OPENGL_ES_BIT,
    'es': EGL_OPENGL_ES_BIT,
}
API_NAMES = dict([
    (k,{
        EGL_OPENGL_BIT:EGL_OPENGL_API,
        EGL_OPENGL_ES2_BIT:EGL_OPENGL_ES_API,
        EGL_OPENGL_ES_BIT:EGL_OPENGL_ES_API
    }[v])
    for k,v in API_BITS.items()
])


def egltest( size=(300,300), name=None, api='es2', attributes=DESIRED_ATTRIBUTES ):
    def gltest( function ):
        """Decorator to allow a function to run in a Pygame GLES[1,2,3] context"""
        @wraps( function )
        def test_function( *args, **named ):
            major,minor = ctypes.c_long(),ctypes.c_long()
            display = eglGetDisplay(EGL_DEFAULT_DISPLAY)
            eglInitialize( display, major, minor)
            num_configs = ctypes.c_long()
            configs = (EGLConfig*2)()
            api_constant = API_NAMES[api.lower()]
            local_attributes = attributes[:]
            local_attributes.extend( [
                EGL_CONFORMANT, API_BITS[api.lower()],
                EGL_NONE,
            ])
            print('local_attributes', local_attributes)
            local_attributes= arrays.GLintArray.asArray( local_attributes )
            eglChooseConfig(display, local_attributes, configs, 2, num_configs)
            print('API', api_constant)
            eglBindAPI(api_constant)
            
            # now need to get a raw X window handle...
            pygame.init()
            pygame.display.set_mode( size )
            window = pygame.display.get_wm_info()['window']
            surface = eglCreateWindowSurface(display, configs[0], window, None )
            
            ctx = eglCreateContext(display, configs[0], EGL_NO_CONTEXT, None)
            if ctx == EGL_NO_CONTEXT:
                raise RuntimeError( 'Unable to create context' )
            try:
                eglMakeCurrent( display, surface, surface, ctx )
                function(*args, **named)
                eglSwapBuffers( display, surface )
            finally:
                pygame.display.quit()
                pygame.quit()
        return test_function
    return gltest 



from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *
glutInit([''])
glutInitDisplayMode (GLUT_SINGLE | GLUT_RGB)
try:
    if fgDeinitialize: fgDeinitialize(False)
except NameError as err:
    pass # Older PyOpenGL, you may see a seg-fault here...



import logging, os, urllib, traceback, textwrap, keyword
import xmlreg
from OpenGL._bytes import as_8_bit, as_str, unicode, as_unicode
HERE = os.path.join( os.path.dirname(__file__))
log = logging.getLogger( __name__ )
AUTOGENERATION_SENTINEL = """### DO NOT EDIT above the line "END AUTOGENERATED SECTION" below!"""
AUTOGENERATION_SENTINEL_END = """### END AUTOGENERATED SECTION"""

def nameToPathMinusGL( name ):
    return "/".join( name.split( '_',2 )[1:] )
def prefix_split(name):
    return name.split( '_',2 )[1:]
def indent( text, indent='\t' ):
    return "\n".join([
        '%s%s'%(indent,line) 
        for line in text.splitlines()
    ])

class Generator( object ):
    targetDirectory = os.path.join( HERE, '..','OpenGL')
    rawTargetDirectory = os.path.join( HERE,'..','OpenGL','raw')
    prefix = 'GL'
    dll = '_p.PLATFORM.GL'
    includeOverviews = True
    
    def __init__( self, registry, type_translator ):
        self.registry = registry
        self.type_translator = type_translator
    def module( self, module ):
        if hasattr( module, 'apis' ):
            for api in module.apis:
                if api != 'glcore':
                    gen = ModuleGenerator(module,self, api)
                    gen.generate()
        else:
            gen = ModuleGenerator(module,self)
            gen.generate()
        return gen
    GLGET_PARAM_GROUPS = [
        #'MaterialParameter',
        #'PixelMap',
        #'LightParameter',
        'GetPName',
        #'GetPixelMap',
        #'GetMapQuery',
        'GetPointervPName',
        #'TextureEnvParameter',
        #'TextureGenParameter',
    ]
    GL_GET_TEMPLATE = '''"""glGet* auto-generation of output arrays (DO NOT EDIT, AUTOGENERATED)"""
try:
    from OpenGL.raw.%(prefix)s._lookupint import LookupInt as _L
except ImportError:
    def _L(*args):
        raise RuntimeError( "Need to define a lookupint for this api" )
_glget_size_mapping = _m = {}
%(elements)s
'''
    def group_sizes( self ):
        """Generate a group-sizes data-table for the given group-name"""
        result = []
        for enum_name, size in sorted( self.glGetSizes.items(), key = lambda x: (bool(x[1]),x[0]) ):
            value = self.registry.enumeration_set.get( enum_name )
            size = [x for x in size if x]
            comment = ''
            if not size:
                size = 'TODO'
                comment = '# '
            else:
                size = ''.join( size )
            if value is None:
                # common in cases where GL and GLES constants are updated together...
                log.debug( 'Unrecognized constant: %s in GLGet section', enum_name )
            else:
                value = value.value
                result.append( 
                    '%(comment)s_m[%(value)s] = %(size)s # %(enum_name)s'%locals()
                )
        elements = '\n'.join( result )
        prefix = self.prefix
        return self.GL_GET_TEMPLATE%locals()
    def enum( self, enum ):
        comment = ''
        try:
            int( enum.value, 0 )
        except ValueError:
            comment = '# '
        return '%s%s=_C(%r,%s)'%(comment, enum.name,enum.name,enum.value)
    def safe_name( self, name ):
        if keyword.iskeyword( name ):
            return name + '_'
        return name
    def function( self, function ):
        """Produce a declaration for this function in ctypes format"""
        returnType = self.type_translator( function.returnType )
        if returnType == 'arrays.GLbyteArray':
            returnType = 'ctypes.c_char_p'
        if function.argTypes:
            argTypes = ','.join([self.type_translator(x) for x in function.argTypes])
        else:
            argTypes = ''
        if function.argNames:
            argNames = ','.join([self.safe_name(n) for n in function.argNames])
        else:
            argNames = ''
        arguments = ', '.join([
            '%s %s'%(t,self.safe_name(n))
            for (t,n) in zip( function.argTypes,function.argNames )
        ])
        name = function.name 
        if returnType.strip() in ('_cs.GLvoid', '_cs.void','void'):
            returnType = pyReturn = 'None'
        else:
            pyReturn = function.returnType
        doc = '%(name)s(%(arguments)s) -> %(pyReturn)s'%locals()
#        log.info( '%s', doc )
        formatted=  self.FUNCTION_TEMPLATE%locals()
        return formatted
    FUNCTION_TEMPLATE = """@_f
@_p.types(%(returnType)s,%(argTypes)s)
def %(name)s(%(argNames)s):pass"""

    _glGetSizes = None
    @property
    def glGetSizes( self ):
        if self._glGetSizes is None:
            self._glGetSizes = self.loadGLGetSizes()
        return self._glGetSizes
    def loadGLGetSizes( self ):
        """Load manually-generated table of glGet* sizes"""
        table = {}
        try:
            lines = [
                line.split('\t')
                for line in open( os.path.join( HERE, 'glgetsizes.csv') ).read().splitlines()
            ]
        except IOError:
            pass 
        else:
            for line in lines:
                if line and line[0]:
                    value = [
                        v for v in [
                            v.strip('"') for v in line[1:]
                        ]
                        if v
                    ]
                    if value:
                        table[line[0].strip('"').strip()] = value
        # now make sure everything registered in the xml file is present...
        output_group_names = {}
        for function in self.registry.command_set.values():
            output_group_names.update( function.output_groups )
        for output_group in output_group_names.keys():
            log.debug( 'Output parameter group: %s', output_group )
            for name in self.registry.enum_groups.get(output_group,[]):
                if name not in table:
                    log.info( 'New %s value: %r', output_group, name )
                    table[name] = ''
        return table
    def saveGLGetSizes( self ):
        """Save out sorted list of glGet sizes to disk"""
        items = self.glGetSizes.items()
        items.sort()
        data = "\n".join([
            '%s\t%s'%(
                key,"\t".join(value)
            )
            for (key,value) in items 
        ])
        open( os.path.join( HERE, 'glgetsizes.csv'),'w').write( data )

class ModuleGenerator( object ):
    ROOT_EXTENSION_SOURCE = 'http://www.opengl.org/registry/specs/'
    RAW_MODULE_TEMPLATE = """'''Autogenerated by xml_generate script, do not edit!'''
from OpenGL import platform as _p, arrays
# Code generation uses this
from OpenGL.raw.%(prefix)s import _types as _cs
# End users want this...
from OpenGL.raw.%(prefix)s._types import *
from OpenGL.raw.%(prefix)s import _errors
from OpenGL.constant import Constant as _C
%(extra_imports)s
import ctypes
_EXTENSION_NAME = %(constantModule)r
def _f( function ):
    return _p.createFunction( function,%(dll)s,%(constantModule)r,error_checker=_errors._error_checker)
%(constants)s
%(declarations)s
"""

    INIT_TEMPLATE = """
def glInit%(camelModule)s%(owner)s():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )
"""
    FINAL_MODULE_TEMPLATE = """'''OpenGL extension %(owner)s.%(module)s

This module customises the behaviour of the 
OpenGL.raw.%(prefix)s.%(owner)s.%(module)s to provide a more 
Python-friendly API

%(overview)sThe official definition of this extension is available here:
%(ROOT_EXTENSION_SOURCE)s%(owner)s/%(module)s.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.%(prefix)s import _types, _glgets
from OpenGL.raw.%(prefix)s.%(owner)s.%(module)s import *
from OpenGL.raw.%(prefix)s.%(owner)s.%(module)s import _EXTENSION_NAME
%(init_function)s
%(output_wrapping)s
"""
    dll = '_p.PLATFORM.GL'
    def __init__( self, registry, overall, api=None ):
        self.registry = registry 
        self.overall = overall
        name = registry.name
        if name in ('GL_ES_VERSION_3_1','GL_ES_VERSION_3_0'):
            api = 'gles3'
            name = 'GLES3'+name[5:]
        if api:
            self.prefix = api.upper()
        else:
            if hasattr( self.registry, 'api' ):
                self.prefix = self.registry.api.upper()
            else:
                self.prefix = name.split('_')[0]
        name = name.split('_',1)[1]
        try:
            self.owner, self.module = name.split('_',1)
            self.sentinelConstant = '%s_%s'%(self.owner,self.module)
            
        except ValueError:
            if name.endswith( 'SGIX' ):
                self.prefix = "GL"
                self.owner = 'SGIX'
                self.module = name[3:-4]
                self.sentinelConstant = '%s%s'%(self.module,self.owner)
            else:
                log.error( """Unable to parse module name: %s""", name )
                raise
        self.dll = '_p.PLATFORM.%s'%(self.prefix,)
        if self.module[0].isdigit():
            self.module = '%s_%s'%(self.prefix,self.module,)
        if self.module == 'async':
            self.module = 'async_'
        self.camelModule = "".join([x.title() for x in self.module.split('_')])
        self.rawModule = self.module
        
        self.rawOwner = self.owner
        while self.owner and self.owner[0].isdigit():
            self.owner = self.owner[1:]
        self.rawPathName = os.path.join( self.overall.rawTargetDirectory, self.prefix, self.owner, self.module+'.py' )
        self.pathName = os.path.join( self.overall.targetDirectory, self.prefix, self.owner, self.module+'.py' )
        
        self.constantModule = '%(prefix)s_%(owner)s_%(rawModule)s'%self
        specification = self.getSpecification()
        self.constantsFromSpec()
        self.overview = ''
        if self.overall.includeOverviews:
            for title,section in specification.blocks( specification.source ):
                if title.startswith( 'Overview' ):
                    if isinstance(section,bytes):
                        section = section.decode('latin-1') # seems not to be utf in at lest some cases
                    self.overview = 'Overview (from the spec)\n%s\n\n'%(
                        indent( 
                            as_unicode(
                                as_str(section)
                                .replace('\xd4','O')
                                .replace('\xd5','O')
                            ).encode( 'ascii', 'ignore' )
                            .decode( 'ascii', 'ignore' )
                        )
                    )
                    break
    def __getitem__( self, key ):
        try:
            return getattr( self, key )
        except AttributeError:
            raise KeyError( key )
    def __getattr__( self, key ):
        if key not in ('registry',):
            return getattr( self.registry, key )
    @property
    def extra_imports( self ):
        if self.name == 'GL_VERSION_1_1':
            # spec files have not properly separated out these two...
            return '# Spec mixes constants from 1.0 and 1.1\nfrom OpenGL.raw.GL.VERSION.GL_1_0 import *'
        return ''
    
    def constantsFromSpec(self, spec=None):
        """Examine spec text looking for new constants..."""
        spec = spec or self.getSpecification()
        if spec.source:
            table = self.overall.glGetSizes
            extras = spec.glGetConstants()
            changed = False
            for key, value in extras.items():
                if key not in table:
                    try:
                        value = int(value, 16)
                    except ValueError:
                        continue 
                    short_key = '_'.join(key.split("_")[:-1])
                    if short_key in table:
                        short_def = self.overall.registry.enumeration_set.get(short_key)
                        if short_def and int(short_def.value, 16) == value:
                            key = short_def.name 
                    if not key in table:
                        table[key] = ['(1,)', '#TODO Review %s'%(self.specFile())]
                        changed = True
            if changed:
                self.overall.saveGLGetSizes()
    def shouldReplace( self ):
        """Should we replace the given filename?"""
        filename = self.pathName
        if not os.path.isfile(
            filename
        ):
            return True
        else:
            hasLines = 0
            for line in open( filename ):
                if line.strip() == AUTOGENERATION_SENTINEL_END.strip():
                    return True
                hasLines = 1
            if not hasLines:
                return True
            log.warning( 'Not replacing %s (no AUTOGENERATION_SENTINEL_END found)', filename )
        return False
    @property
    def output_wrapping( self ):
        """Generate output wrapping statements for our various functions"""
        try:
            statements = []
            for function in self.registry.commands():
                dependencies = function.size_dependencies
                if dependencies: # temporarily just do single-output functions...
                    base = []
                    for param,dependency in sorted(dependencies.items()):
                        param = as_str( param )
                        if isinstance( dependency, xmlreg.Output ):
                            statements.append( '# %s.%s is OUTPUT without known output size'%(
                                function.name,param,
                            ))
                        if isinstance( dependency, xmlreg.Staticsize ):
                            base.append( '.setOutput(\n    %(param)r,size=(%(dependency)r,),orPassIn=True\n)'%locals())
                        elif isinstance( dependency, xmlreg.Dynamicsize ):
                            base.append( '.setOutput(\n    %(param)r,size=lambda x:(x,),pnameArg=%(dependency)r,orPassIn=True\n)'%locals())
                        elif isinstance( dependency, xmlreg.Multiple ):
                            pname,multiple = dependency
                            base.append( '.setOutput(\n    %(param)r,size=lambda x:(x,%(multiple)s),pnameArg=%(pname)r,orPassIn=True\n)'%locals())
                        elif isinstance( dependency, xmlreg.Compsize ):
                            if len(dependency) == 1:
                                pname = dependency[0]
                                base.append( '.setOutput(\n    %(param)r,size=_glgets._glget_size_mapping,pnameArg=%(pname)r,orPassIn=True\n)'%locals())
                            else:
                                statements.append('# OUTPUT %s.%s COMPSIZE(%s) '%(function.name,param,', '.join(dependency)) )
                        elif isinstance( dependency, xmlreg.StaticInput ):
                            base.append( '.setInputArraySize(\n    %(param)r, %(dependency)s\n)'%locals())
                        elif isinstance( dependency, (xmlreg.DynamicInput, xmlreg.MultipleInput, xmlreg.Input) ):
                            if dependency is None:
                                continue 
                            statements.append( '# INPUT %s.%s size not checked against %s'%(
                                function.name, 
                                param,
                                dependency
                            ))
                            base.append( '.setInputArraySize(\n    %(param)r, None\n)'%locals())
                    if base:
                        base.insert(0, '%s=wrapper.wrapper(%s)'%(function.name,function.name) )
                        statements.append( ''.join(base ))
            return '\n'.join( statements )
        except Exception as err:
            traceback.print_exc()
            import pdb 
            pdb.set_trace()
    
    def get_constants( self ):
        functions = self.registry.enums()
        functions.sort( key = lambda x: x.name )
        return functions
    @property 
    def init_function( self ):
        return self.INIT_TEMPLATE%self
    @property
    def constants( self ):
        result = []
        try:
            for function in self.get_constants():
                result.append( self.overall.enum( function ) )
            return '\n'.join(result)
        except Exception:
            traceback.print_exc()
            raise
    @property
    def declarations( self ):
        functions = self.registry.commands()
        functions.sort( key = lambda x: x.name )
        result = []
        for function in functions:
            result.append( self.overall.function( function ) )
        return "\n".join( result )
    EXTENSION_BASE = os.path.join(HERE,'khronosapi','extensions')
    def specFile(self):
        """Lookup the specification file for our extension"""
        prefix,name = prefix_split(self.name)
        specFile = os.path.join(self.EXTENSION_BASE,prefix,'%s_%s.txt'%( 
            prefix,name,
        ))
        return specFile
    def getSpecification( self ):
        """Retrieve our specification document...
        
        Retrieves the .txt file which defines this specification,
        allowing us to review the document locally in order to provide
        a reasonable wrapping of it...
        """
        if self.registry.feature:
            return Specification('')
        specFile = self.specFile()
        if os.path.exists(specFile):
            data = open( specFile,'rb' ).read()
            return Specification( data )
        else:
            log.info("No spec-file in %s",specFile)
            return Specification('')
    
    def generate( self ):
        for target in (self.rawPathName,self.pathName):
            directory = os.path.dirname( target )
            if not os.path.exists( directory ):
                log.warning( 'Creating target directory: %s', directory )
                os.makedirs( directory )
            if not os.path.isfile( os.path.join(directory, '__init__.py')):
                open( os.path.join(directory, '__init__.py'),'w').write( 
                    '''"""OpenGL Extensions"""'''
                )
            
        directory = os.path.dirname(self.rawPathName)
        current = ''
        toWrite = self.RAW_MODULE_TEMPLATE % self
        try:
            current = open( self.rawPathName, 'r').read()
        except Exception as err:
            pass 
        if current.strip() != toWrite.strip():
            fh = open( self.rawPathName, 'w')
            fh.write( toWrite )
            fh.close()
        if isinstance( self.registry, xmlreg.Feature ):
            # this is a core feature...
            target = os.path.join( self.overall.rawTargetDirectory, self.prefix,'_glgets.py' )
            open( target,'w' ).write( self.overall.group_sizes())
        if self.shouldReplace( ):
            # now the final module with any included custom code...
            toWrite = self.FINAL_MODULE_TEMPLATE % self
            current = ''
            try:
                current = open( self.pathName, 'r').read()
            except Exception as err:
                pass 
            else:
                found = current.rfind( '\n'+AUTOGENERATION_SENTINEL_END )
                if found >= -1:
                    if current[:found].strip() == toWrite.strip():
                        # we aren't going to change anything...
                        return False
                    found += len( '\n' + AUTOGENERATION_SENTINEL_END )
                    current = current[found:]
                else:
                    current = ''
            try:
                fh = open( self.pathName, 'w')
            except IOError as err:
                log.warning( "Unable to create module for %r %s", self.name, err )
                return False
            else:
                fh.write( toWrite )
                fh.write( AUTOGENERATION_SENTINEL_END )
                fh.write( current )
                fh.close()
                return True
        return False

class Specification( object ):
    """Parser for parsing OpenGL specifications for interesting information
    """
    def __init__( self, source ):
        """Store the source text for the specification"""
        if isinstance(source,bytes):
            try:
                self.source = source.decode('utf-8')
            except UnicodeDecodeError as err:
                self.source = source.decode('latin-1')
        else:
            self.source = source
        assert isinstance(self.source, unicode)
    def blocks( self, data ):
        """Retrieve the set of all blocks"""
        data = data.splitlines()
        title = []
        block = []
        for line in data:
            if line and line.lstrip() == line:
                if block:
                    yield "\n".join(title), textwrap.dedent( "\n".join(block) )
                    title = [ ]
                    block = [ ]
                title.append( line )
            else:
                block.append( line )
        if block:
            yield "\n".join(title), textwrap.dedent( "\n".join(block) )
    def constantBlocks( self ):
        """Retrieve the set of constant blocks"""
        for title,block in self.blocks( self.source ):
            if title and title.startswith( 'New Tokens' ):
                yield block
    def glGetConstants( self ):
        """Retrieve the set of constants which pass to glGet* functions"""
        table = {}
        for block in self.constantBlocks():
            for title, section in self.blocks( block ):
                for possible in (
                    'GetBooleanv','GetIntegerv','<pname> of Get', 
                    '<pname> parameter of Get', 
                ):
                    if possible in title:
                        for line in section.splitlines():
                            line = line.strip().split()
                            if len(line) == 2:
                                constant,value = line 
                                constant = constant.strip(':')
                                table['GL_%s'%(constant,)] = value 
                        break
        return table

def download( url ):
    """Download the given url, informing the user of what we're doing"""
    log.info( 'Download: %r',url,)
    file = urllib.urlopen( url )
    return file.read()



from OpenGL.GLUT import *
glutInit()
glutInitDisplayMode(GLUT_RGB)
try:
    if fgDeinitialize: fgDeinitialize(False)
except NameError as err:
    pass # Older PyOpenGL, you may see a seg-fault here...



#! /usr/bin/env python
import logging
import os_egl
from OpenGL.EGL import *
from OpenGL.EGL.EXT.device_enumeration import eglQueryDevicesEXT
log = logging.getLogger(__name__)

def main():
    devices = (EGLDeviceEXT * 10)()
    count = EGLint()
    major, minor = EGLint(), EGLint()
    if eglQueryDevicesEXT(10,devices,count):
        log.info("%s devices enumerated", count)
        for i,device in enumerate(devices[:count.value]):
            display = eglGetDisplay(device)
            if not eglInitialize(display,major,minor):
                log.info("Could not initialise, skipping")
                continue
            log.info("Display #%d, Version %s.%s",i,major.value,minor.value)
            for key in [
                EGL_VENDOR,
            ]:
                log.info("#%d %s", i, eglQueryString(display,key))
#        display = eglGetDisplay(EGL_DEFAULT_DISPLAY)

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    main()



#! /usr/bin/env python
"""Script to pull out OpenGL core definitions from the linux opengl header"""
import re
import get_gl_extensions
constant = re.compile( '^[#]define\W+(?P<name>[A-Z_0-9]+)\W+(?P<value>0x[a-fA-F0-9]+)$', re.M )
function = re.compile( '^GLAPI\W+(?P<return>[A-Z_0-9a-z *]+)\W+GLAPIENTRY\W+(?P<name>[a-zA-Z_0-9]+)(?P<signature>[(].*?[)])\W*[;]', re.DOTALL|re.M )

class CoreModule( get_gl_extensions.VersionModule ):
    _header = None
    dll = 'None'
    GLGET_CONSTANT = '_GLGET_CONSTANTS[ %(name)s ] = %(size)s'
    @property
    def RAW_MODULE_TEMPLATE( self ):
        assert not 'glget' in get_gl_extensions.WRAPPER_TEMPLATE
        return get_gl_extensions.WRAPPER_TEMPLATE + get_gl_extensions.INIT_TEMPLATE
    def read_header( self ):
        if not self._header:
            content = open( '/usr/include/GL/gl.h' ).read()
            start = content.index( ' * Constants' )
            stop = content.index( ' * OpenGL 1.2' )
            self._header = content[start:stop]
        return self._header

    def findFunctions( self ):
        declarations = self.read_header()
        functions = []
        for match in function.finditer(declarations):
            match_dict = match.groupdict()
            f = get_gl_extensions.Function( match_dict['return'], match_dict['name'], " ".join(match_dict['signature'].splitlines()) )
            if 'ptr' in f.argNames:
                f.argNames[f.argNames.index('ptr')] = 'pointer'
            functions.append( f )
        self.functions = functions

def core():
    # avoid import loops...
    get_gl_extensions.WRAPPER_TEMPLATE = get_gl_extensions.WRAPPER_TEMPLATE.replace( 
        'from OpenGL.GL import glget', '_GLGET_CONSTANTS = {}', 
    )
    assert not 'glget' in get_gl_extensions.WRAPPER_TEMPLATE
    class FakeHeader( object ):
        includeOverviews = False
        registry = {}
        glGetSizes = {}
        def loadGLGetSizes( self ):
            """Load manually-generated table of glGet* sizes"""
            table = self.glGetSizes
            try:
                lines = [
                    line.split('\t')
                    for line in open( 'glgetsizes.csv' ).read().splitlines()
                ]
            except IOError, err:
                pass 
            else:
                for line in lines:
                    if line and line[0]:
                        table[line[0].strip('"')] = [
                            v for v in [
                                v.strip('"') for v in line[1:]
                            ]
                            if v
                        ]
    f = FakeHeader()
    f.loadGLGetSizes()
    c = CoreModule( 'GL_VERSION_GL_1_1', [], f )
    c.segments = [c.read_header()]
    c.process()

if __name__ == "__main__":
    core()



from __future__ import print_function
#import OpenGL
#OpenGL.USE_ACCELERATE=False
from OpenGL.GL import *
import pytest
import sys
if not sys.platform.startswith('linux'):
    pytest.skip("Skipping GLX tests on non-linux platforms", allow_module_level=True)
from OpenGL.GLX import *
from OpenGL.GLX.EXT.texture_from_pixmap import *
from pygamegltest import pygametest
import os

attributes = [
#    GLX_BIND_TO_TEXTURE_RGBA_EXT, 1,
#    GLX_DRAWABLE_TYPE, GLX_PIXMAP_BIT,
#    GLX_BIND_TO_TEXTURE_TARGETS_EXT, GLX_TEXTURE_2D_BIT_EXT,
    GLX_DOUBLEBUFFER, 1,
#    GLX_Y_INVERTED_EXT, GLX_DONT_CARE,
    GL_NONE
]

#from OpenGL import platform
import ctypes
from OpenGL.platform import ctypesloader

X11 = ctypesloader.loadLibrary( ctypes.cdll, 'X11' )
XDefaultScreen = X11.XDefaultScreen
XDefaultScreen.argtypes = [ctypes.POINTER(Display)]
XOpenDisplay = X11.XOpenDisplay 
XOpenDisplay.restype = ctypes.POINTER(Display)

@pygametest()
def main():
    dsp = XOpenDisplay( os.environ.get( 'DISPLAY' ))
    screen = XDefaultScreen( dsp )
    print('X Display %s Screen %s'%( dsp, screen ))
    major,minor = GLint(),GLint()
    glXQueryVersion(dsp, major, minor)
    version = (major.value,minor.value)
    print('glX Version: %s.%s'%version)
    if version >= (1,1):
        print(glXQueryExtensionsString(dsp,screen))
        if version >= (1,2):
            d = glXGetCurrentDisplay()[0]
            print('Current display', d)
        else:
            d = dsp
    if version >= (1,3):
        elements = GLint(0)
        configs = glXChooseFBConfig(
            dsp, 
            screen, 
            (GLint * len(attributes))( * attributes ), 
            elements
        )
        print('%s configs found'%( elements.value ))
        for config in range( elements.value ):
            print('Config: %s %s'%(config,configs[config][0]))
            samples = ctypes.c_int()
            for attribute in (
                'GLX_FBCONFIG_ID','GLX_BUFFER_SIZE',
                'GLX_LEVEL','GLX_DOUBLEBUFFER',
                'GLX_STEREO',
                'GLX_SAMPLES','GLX_SAMPLE_BUFFERS',
                'GLX_DRAWABLE_TYPE',
            ):
                glXGetFBConfigAttrib( dsp, configs[config], globals()[attribute], samples )
                print('%s -> %s'%( attribute, samples.value ))
            print() 
    from OpenGL.raw.GLX import _types
    print('Extension List', _types.GLXQuerier.getExtensions())
        
if __name__ == "__main__":
    main()



#! /usr/bin/env python
from __future__ import print_function
import pygame, pygame.display
import logging, time, traceback, unittest, os
logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)
HERE = os.path.dirname( __file__ )
import pickle
try:
    import cPickle
except ImportError as err:
    cPickle = pickle

try:
    from numpy import *
except ImportError as err:
    array = None

pygame.display.init()
import OpenGL 
if os.environ.get( 'TEST_NO_ACCELERATE' ):
    OpenGL.USE_ACCELERATE = False
#OpenGL.FULL_LOGGING = True
OpenGL.CONTEXT_CHECKING = True
OpenGL.FORWARD_COMPATIBLE_ONLY = False
OpenGL.UNSIGNED_BYTE_IMAGES_AS_STRING = True

#from OpenGL._bytes import bytes, _NULL_8_BYTE, unicode, as_8_bit
from OpenGL.GL import *
try:
    glGetError()
except error.NoContext as err:
    # good, should have got this error 
    pass
else:
    print( 'WARNING: Failed to catch invalid context' )
    #raise RuntimeError( """Did not catch invalid context!""" )
from OpenGL import error
from OpenGL.GLU import *
import OpenGL
from OpenGL.extensions import alternate, GLQuerier
from OpenGL.GL.framebufferobjects import *
from OpenGL.GL.EXT.multi_draw_arrays import *
from OpenGL.GL.ARB.imaging import *
from OpenGL._bytes import _NULL_8_BYTE

glMultiDrawElements = alternate( 
    glMultiDrawElementsEXT, glMultiDrawElements, 
)
import basetestcase

class TestCore( basetestcase.BaseTest ):
    def test_errors( self ):
        """Test for error catching/checking"""
        try:
            glClear( GL_INVALID_VALUE )
        except Exception as err:
            assert err.err == 1281, ("""Expected invalid value (1281)""", err.err)
        else:
            assert not OpenGL.ERROR_CHECKING, """No error on invalid glClear"""
        try:
            glColorPointer(GL_INVALID_VALUE,GL_BYTE,0,None)
        except Exception as err:
            assert err.err == 1281, ("""Expected invalid value (1281)""", err.err)
            assert err.baseOperation, err.baseOperation
            assert err.pyArgs == (GL_INVALID_VALUE, GL_BYTE, 0, None), err.pyArgs
            assert err.cArgs == (GL_INVALID_VALUE, GL_BYTE, 0, None), err.cArgs
        else:
            assert not OpenGL.ERROR_CHECKING, """No error on invalid glColorPointer"""
        try:
            glBitmap(-1,-1,0,0,0,0,as_8_bit(""))
        except Exception as err:
            assert err.err in (1281,1282), ("""Expected invalid value (1281) or invalid operation (1282)""", err.err)
        else:
            assert not OpenGL.ERROR_CHECKING, """No error on invalid glBitmap"""
    if not OpenGL.ERROR_ON_COPY:
        def test_simple( self ):
            """Test for simple vertex-based drawing"""
            glDisable( GL_LIGHTING )
            glBegin( GL_TRIANGLES )
            try:
                try:
                    glVertex3f( 0.,1.,0. )
                except Exception:
                    traceback.print_exc()
                glVertex3fv( [-1,0,0] )
                glVertex3dv( [1,0,0] )
                try:
                    glVertex3dv( [1,0,4,5] )
                except ValueError:
                    #Got expected value error (good)
                    assert OpenGL.ARRAY_SIZE_CHECKING, """Should have raised ValueError when doing array size checking"""
                else:
                    assert not OpenGL.ARRAY_SIZE_CHECKING, """Should not have raised ValueError when not doing array size checking"""
            finally:
                glEnd()
            a = glGenTextures( 1 )
            assert a
            b = glGenTextures( 2 )
            assert len(b) == 2
    def test_arbwindowpos( self ):
        """Test the ARB window_pos extension will load if available"""
        from OpenGL.GL.ARB.window_pos import glWindowPos2dARB
        if glWindowPos2dARB:
            glWindowPos2dARB( 0.0, 3.0 )
    def test_getstring( self ):
        assert glGetString( GL_EXTENSIONS )
    def test_constantPickle( self ):
        """Test that our constants can be pickled/unpickled properly"""
        for p in pickle,cPickle:
            v = p.loads( p.dumps( GL_VERTEX_ARRAY ))
            assert v == GL_VERTEX_ARRAY, (v,GL_VERTEX_ARRAY)
            assert v.name == GL_VERTEX_ARRAY.name, v.name 
    
    
    def test_nonFloatColor( self ):
        """Test that we handle non-floating-point colour inputs"""
        for notFloat,shouldWork in ((0,True), (object(),False), (object,False)):
            try:
                glColor4f( 0,1,1,notFloat )
            except Exception:
                if shouldWork:
                    raise 
            else:
                if not shouldWork:
                    raise RuntimeError( """Expected failure for non-float value %s, succeeded"""%( notFloat, ))
    if array:
        def test_arrayTranspose( self ):
            m = glGetFloatv( GL_MODELVIEW_MATRIX )
            glMatrixMode( GL_MODELVIEW )
            glLoadIdentity()

            t = eye(4)
            t[3,0] = 20.0

            # the following glMultMatrixf call ignored this transpose
            t = t.T
            if OpenGL.ERROR_ON_COPY:
                t = ascontiguousarray( t )
            
            glMultMatrixd( t )

            m = glGetFloatv( GL_MODELVIEW_MATRIX )
            assert allclose( m[-1], [0,0,0,1] ), m
    
    if array:
        # currently crashes in py_buffer operation, so reverted to raw numpy 
        # api
        def test_mmap_data( self ):
            """Test that we can use mmap data array
            
            If we had a reasonable lib that dumped raw image data to a shared-mem file
            we might be able to use this for movie display :) 
            """
            fh = open( 'mmap-test-data.dat', 'wb+' )
            fh.write( _NULL_8_BYTE*(32*32*3+1))
            fh.flush()
            fh.close()
            # using memmap here...
            data = memmap( 'mmap-test-data.dat' )
            for i in range( 0,255,2 ):
                glDrawPixels( 32,32, GL_RGB, GL_UNSIGNED_BYTE, data, )
                glFlush()
                pygame.display.flip()
                data[::2] = i
                time.sleep( 0.001 )
    
    if array:
        def test_vbo( self ):
            """Test utility vbo wrapper"""
            from OpenGL.arrays import vbo
            if not vbo.get_implementation():
                return
            if not glVertexPointerd or not glDrawElements:
                return
            points = array( [
                [0,0,0],
                [0,1,0],
                [1,.5,0],
                [1,0,0],
                [1.5,.5,0],
                [1.5,0,0],
            ], dtype='d')
            indices = array(
                range(len(points)),
                ['i','I'][bool(OpenGL.ERROR_ON_COPY)], # test coercion if we can
            )
            d = vbo.VBO(points)
            glDisable( GL_CULL_FACE )
            glNormal3f( 0,0,1 )
            glColor3f( 1,1,1 )
            glEnableClientState(GL_VERTEX_ARRAY)
            try:
                for x in range( 1, 255, 10 ):
                    d.bind()
                    try:
                        glVertexPointerd( d )
                        glDrawElements( GL_LINE_LOOP, len(indices), GL_UNSIGNED_INT, indices )
                    finally:
                        d.unbind()
                    lastPoint = array( [[1.5,(1/255.) * float(x),0]] )
                    d[-2:-1] = lastPoint
                    glFlush()
                    pygame.display.flip()
                    glClear( GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT )
                    time.sleep( 0.001 )
            finally:
                glDisableClientState( GL_VERTEX_ARRAY )
            # bug report from Dan Helfman, delete shouldn't cause 
            # errors if called explicitly
            d.delete()
        def test_glgetbufferparameter(self):
            if not glGenBuffers or not glGenVertexArrays:
                return None
            buffer = glGenBuffers(1)
            vertex_array = glGenVertexArrays(1,buffer)
            glBindBuffer(GL_ARRAY_BUFFER, buffer)
            try:
                mapped = glGetBufferParameteriv(GL_ARRAY_BUFFER, GL_BUFFER_MAPPED)
                assert mapped == (GL_FALSE if OpenGL.SIZE_1_ARRAY_UNPACK else [GL_FALSE]), mapped
            finally:
                glBindBuffer(GL_ARRAY_BUFFER, 0)
                glDeleteVertexArrays(1,vertex_array)
                glDeleteBuffers(1,buffer)
    def test_glbufferparameter_create(self):
        if not glGenBuffers or not glGenVertexArrays:
            return None
        for create in [True,False]:
            buffer = glGenBuffers(1)
            vertex_array = glGenVertexArrays(1,buffer)
            glBindBuffer(GL_ARRAY_BUFFER, buffer)
            try:
                for param, expected in [
                    (GL_BUFFER_SIZE,0),
                    (GL_BUFFER_MAPPED,GL_FALSE),
                    (GL_BUFFER_STORAGE_FLAGS,0),
                    (GL_BUFFER_USAGE,GL_STATIC_DRAW),
                ]:
                    if create:
                        mapped = GLint(-1)
                        glGetBufferParameteriv(GL_ARRAY_BUFFER, param, mapped)
                        assert mapped.value == expected, (param, mapped, expected)
                    else:
                        mapped = glGetBufferParameteriv(GL_ARRAY_BUFFER, param)
                        if param == GL_BUFFER_USAGE:
                            assert mapped[0] == expected, (param, mapped, expected)
                        else:
                            if OpenGL.SIZE_1_ARRAY_UNPACK:
                                assert mapped == expected, (param, mapped, expected)
                            else:
                                assert mapped[0] == expected, (param, mapped[0], expected)
            finally:
                glBindBuffer(GL_ARRAY_BUFFER, 0)
                glDeleteVertexArrays(1,vertex_array)
                glDeleteBuffers(1,buffer)
    def test_fbo( self ):
        """Test that we support framebuffer objects
        
        http://www.gamedev.net/reference/articles/article2331.asp
        """
        if not glGenFramebuffers:
            print( 'No Frame Buffer Support!' )
            return False
        width = height = 128
        fbo = glGenFramebuffers(1)
        glBindFramebuffer(GL_FRAMEBUFFER, fbo)
        depthbuffer = glGenRenderbuffers(1 )
        glBindRenderbuffer(GL_RENDERBUFFER, depthbuffer)
        glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT24, width, height)
        glFramebufferRenderbuffer(
            GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, 
            depthbuffer
        )

        img = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D, img)
        # NOTE: these lines are *key*, without them you'll likely get an unsupported format error,
        # ie. GL_FRAMEBUFFER_UNSUPPORTED
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
        glTexImage2D(
            GL_TEXTURE_2D, 0, GL_RGB8,  
            width, height, 0, GL_RGB, 
            GL_INT, 
            None # no data transferred
        ) 
        glFramebufferTexture2D(
            GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, 
            img, 
            0 # mipmap level, normally 0
        )
        status = glCheckFramebufferStatus(GL_FRAMEBUFFER)
        assert status == GL_FRAMEBUFFER_COMPLETE, status
        glBindFramebuffer(GL_FRAMEBUFFER, fbo)
        glPushAttrib(GL_VIEWPORT_BIT) # viewport is shared with the main context
        try:
            glViewport(0,0,width, height)
            
            # rendering to the texture here...
            glColor3f( 1,0,0 )
            glNormal3f( 0,0,1 )
            glBegin( GL_QUADS )
            for v in [[0,0,0],[0,1,0],[1,1,0],[1,0,0]]:
                glColor3f( *v )
                glVertex3d( *v )
            glEnd()
        finally:
            glPopAttrib(); # restore viewport
        glBindFramebuffer(GL_FRAMEBUFFER, 0) # unbind
        
        glBindTexture(GL_TEXTURE_2D, img)
        
        glEnable( GL_TEXTURE_2D )
        
        # rendering with the texture here...
        glColor3f( 1,1,1 )
        glNormal3f( 0,0,1 )
        glDisable( GL_LIGHTING )
        glBegin( GL_QUADS )
        try:
            for v in [[0,0,0],[0,1,0],[1,1,0],[1,0,0]]:
                glTexCoord2f( *v[:2] )
                glVertex3d( *v )
        finally:
            glEnd()
    def test_gl_1_2_support( self ):
        if glBlendColor:
            glBlendColor( .3, .4, 1.0, .3 )
            print('OpenGL 1.2 support')
    if array:
        def test_glmultidraw( self ):
            """Test that glMultiDrawElements works, uses glDrawElements"""
            if glMultiDrawElements:
                points = array([
                    (i,0,0) for i in range( 8 )
                ] + [
                    (i,1,0) for i in range( 8 )
                ], 'd')
                indices = array([
                    [0,8,9,1, 2,10,11,3,],
                    [4,12,13,5,6,14,15,7],
                ],'B')
                index_pointers = arrays.GLvoidpArray.zeros( (2,))
                index_pointers[0] = arrays.GLbyteArray.dataPointer( indices )
                index_pointers[1] = arrays.GLbyteArray.dataPointer( indices[1] )
                counts = [ len(x) for x in indices ]
                if OpenGL.ERROR_ON_COPY:
                    counts = (GLuint*len(counts))(*counts)
                glEnableClientState( GL_VERTEX_ARRAY )
                glDisableClientState( GL_COLOR_ARRAY )
                glDisableClientState( GL_NORMAL_ARRAY )
                try:
                    glVertexPointerd( points )
                    glDisable( GL_LIGHTING )
                    try:
                        glMultiDrawElements(GL_QUAD_STRIP, counts, GL_UNSIGNED_BYTE, index_pointers, 2)
                    finally:
                        glEnable( GL_LIGHTING )
                finally:
                    glDisableClientState( GL_VERTEX_ARRAY )
            else:
                print('No multi_draw_arrays support')
    def test_glDrawBuffers_list( self ):
        """Test that glDrawBuffers with list argument doesn't crash"""
        if not glDrawBuffers:
            return
        a_type = GLenum*2
        args = a_type(
            GL_COLOR_ATTACHMENT0,
            GL_COLOR_ATTACHMENT1,
        )
        try:
            glDrawBuffers( 2, args )
        except GLError as err:
            assert err.err == GL_INVALID_OPERATION, err
    def test_glDrawBuffers_list_valid( self ):
        """Test that glDrawBuffers with list argument where value is set"""
        if not glGenFramebuffers:
            return
        if not glDrawBuffers:
            return
        previous = glGetIntegerv( GL_READ_BUFFER )
        fbo = glGenFramebuffers(1)
        glBindFramebuffer(GL_FRAMEBUFFER, fbo)
        try:
            img1,img2 = glGenTextures(2)
            for img in img1,img2:
                glBindTexture( GL_TEXTURE_2D, img )
                glTexImage2D(
                    GL_TEXTURE_2D, 0, GL_RGB8,  
                    300, 300, 0, GL_RGB, 
                    GL_INT, 
                    None # no data transferred
                ) 
            

            glFramebufferTexture2D(
                GL_FRAMEBUFFER, 
                GL_COLOR_ATTACHMENT0, 
                GL_TEXTURE_2D, img1, 0
            )
            glFramebufferTexture2D(
                GL_FRAMEBUFFER, 
                GL_COLOR_ATTACHMENT1, 
                GL_TEXTURE_2D, img2, 0
            )
            a_type = GLenum*2
            drawingBuffers = a_type(
                GL_COLOR_ATTACHMENT0, 
                GL_COLOR_ATTACHMENT1,
            )
            glDrawBuffers(2, drawingBuffers )
            try:
                checkFramebufferStatus()
            except error.GLError:
                pass
            else:
                glReadBuffer( GL_COLOR_ATTACHMENT1 )
                pixels = glReadPixels( 0,0, 10,10, GL_RGB, GL_UNSIGNED_BYTE )
                assert len(pixels) == 300, len(pixels)
        finally:
            glBindFramebuffer( GL_FRAMEBUFFER, 0 )
        
        glReadBuffer( previous )
        
    def test_get_version( self ):
        from OpenGL.extensions import hasGLExtension
        if hasGLExtension( 'GL_VERSION_GL_2_0' ):
            assert glShaderSource
            assert glUniform1f
        else:
            assert not glShaderSource
            assert not glUniform1f
    
    def test_lookupint( self ):
        from OpenGL.raw.GL import _lookupint 
        if GLQuerier.pullVersion() < [2]:
            return
        l = _lookupint.LookupInt( GL_NUM_COMPRESSED_TEXTURE_FORMATS, GLint )
        result = int(l)
        if not os.environ.get('TRAVIS'):
            assert result, "No compressed textures on this platform? that seems unlikely"
        else:
            "Travis xvfb doesn't normally have compressed textures, possible upgrade?"
    
    def test_glget( self ):
        """Test that we can run glGet... on registered constants without crashing..."""
        from OpenGL.raw.GL import _glgets
        get_items = sorted(_glgets._glget_size_mapping.items())
        for key,value in get_items:
            # There are glGet values that will cause a crash during cleanup
            # GL_ATOMIC_COUNTER_BUFFER_BINDING 0x92c1 crashes/segfaults
            if key >= 0x8df9 and key <= 0x8e23:
                continue
            if key >= 0x92be and key <= 0x92c9:
                continue
            # print( 'Trying glGetFloatv( 0x%x )'%(key,))
            try:
                result = glGetFloatv( key )
            except error.GLError as err:
                if err.err == GL_INVALID_ENUM:
                    pass
                elif err.err == GL_INVALID_OPERATION:
                    if key == 0x882d: # gl draw buffer 
                        pass
                else:
                    raise 
            else:
                if value == (1,) and OpenGL.SIZE_1_ARRAY_UNPACK:
                    try:
                        result = float(result)
                    except TypeError as err:
                        err.args += (result,key)
                        raise
                else:
                    assert ArrayDatatype.dimensions( result ) == value, (result,ArrayDatatype.dimensions( result ), value)
    def test_max_compute_work_group_invocations(self):
        from OpenGL.extensions import hasGLExtension
        if hasGLExtension( 'GL_ARB_compute_shader' ):
            assert glGetIntegerv( GL_MAX_COMPUTE_WORK_GROUP_INVOCATIONS )
    
    
    def test_glCallLists_twice2( self ):
        """SF#2829309 report that glCallLists doubles operation"""
        glRenderMode (GL_RENDER)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(40.0, 1.0, 1.0, 10.0)
        glMatrixMode (GL_MODELVIEW)
        glLoadIdentity ()
        glTranslatef (0, 0, -3)
        first = glGenLists( 2 )
        second = first+1

        glNewList(first, GL_COMPILE_AND_EXECUTE)
        glInitNames ()
        if not OpenGL.ERROR_ON_COPY:
            glCallLists([second]) # replace with gCallList(2)
        else:
            lists = (GLuint * 1)()
            lists[0] = second
            glCallLists(lists)
        #glCallList(second)
        glEndList ()

        glNewList(second, GL_COMPILE)
        glPushName (1)
        glBegin (GL_POINTS)
        glVertex3f (0, 0, 0)
        glEnd ()
        glEndList ()
        glCallList( second )
        glPopName()
        depth = glGetIntegerv( GL_NAME_STACK_DEPTH )
        assert depth in (0,(0,)), depth # have popped, but even then, were' not in the mode...

        glSelectBuffer (100)
        glRenderMode (GL_SELECT)
        glCallList(1)
        depth = glGetIntegerv( GL_NAME_STACK_DEPTH )
        assert depth in (1,(1,)), depth # should have a single record
        glPopName()
        records = glRenderMode (GL_RENDER)
        # reporter says sees two records, Linux sees none, Win32 sees 1 :(
        assert len(records) == 1, records
    
    
    def test_orinput_handling( self ):
        if not glGenVertexArrays:
            return None
        x = glGenVertexArrays(1)
        x = int(x) # check that we got x as an integer-compatible value
        x2 = GLuint()
        r_value = glGenVertexArrays( 1, x2 )
        assert x2.value, x2.value
        assert r_value
        
        color = glGetFloatv( GL_FOG_COLOR )
        color2 = (GLfloat *4)()
        glGetFloatv( GL_FOG_COLOR, color2 )
        for a,b in zip( color,color2 ):
            assert a==b, (color,color2)
    
    
    def test_get_read_fb_binding( self ):
        if GLQuerier.pullVersion() < [3]:
            return
        glGetInteger(GL_READ_FRAMEBUFFER_BINDING)
    
    def test_shader_compile_string( self ):
        if not glCreateShader:
            return None
        shader = glCreateShader(GL_VERTEX_SHADER)
        
        def glsl_version():
            """Parse GL_SHADING_LANGUAGE_VERSION into [int(major),int(minor)]"""
            version = glGetString( GL_SHADING_LANGUAGE_VERSION )
            version = version.split(as_8_bit(' '))[0]
            version = [int(x) for x in version.split(as_8_bit('.'))[:2]]
            return version 
        if glsl_version() < [3,3]:
            return
        SAMPLE_SHADER = '''#version 330
        void main() { gl_Position = vec4(0,0,0,0);}'''
        if OpenGL.ERROR_ON_COPY:
            SAMPLE_SHADER = as_8_bit( SAMPLE_SHADER )
        glShaderSource(shader, SAMPLE_SHADER)
        glCompileShader(shader)
        if not bool(glGetShaderiv(shader, GL_COMPILE_STATUS)) == True:
            print('Info log:')
            print(glGetShaderInfoLog(shader))
            assert False, """Failed to compile"""
    
    def test_gen_framebuffers_twice( self ):
        if not glGenFramebuffersEXT:
            return
        glGenFramebuffersEXT(1)
        f1 = glGenFramebuffersEXT(1)
        f2 = glGenFramebuffersEXT(1)
        assert f1 != f2, (f1,f2)
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, f2)
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0)
    
    def test_compressed_data(self):
        from OpenGL.extensions import hasGLExtension
        if hasGLExtension( 'GL_EXT_texture_compression_s3tc' ):
            from OpenGL.GL.EXT import texture_compression_s3tc as s3tc
            texture = glGenTextures(1)
            glEnable( GL_TEXTURE_2D )
            image_type = GLubyte *256*256
            image = image_type()
            glCompressedTexImage2D(
                GL_TEXTURE_2D, 0, 
                s3tc.GL_COMPRESSED_RGBA_S3TC_DXT5_EXT, 
                256, 256, 0, 
                # 256*256*1,
                image,
            )
            assert texture
    
    
        
if __name__ == "__main__":
    logging.basicConfig( level=logging.INFO )
    unittest.main()
    pygame.display.quit()
    pygame.quit()



"""Test simple functions (i.e. no pointers involved)"""
from __future__ import print_function
from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *
from OpenGL._bytes import as_8_bit
import time
start = time.time()

window = None

def drawText( value, x,y,  windowHeight, windowWidth, step = 18 ):
    """Draw the given text at given 2D position in window
    """
    glMatrixMode(GL_PROJECTION);
    # For some reason the GL_PROJECTION_MATRIX is overflowing with a single push!
    # glPushMatrix()
    matrix = glGetDouble( GL_PROJECTION_MATRIX )
    
    glLoadIdentity();
    glOrtho(0.0, windowHeight or 32, 0.0, windowWidth or 32, -1.0, 1.0)
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glRasterPos2i(x, y);
    lines = 0
    for character in value:
        if character == '\n':
            glRasterPos2i(x, y-(lines*18))
        else:
            glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, ord(character));
    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    # For some reason the GL_PROJECTION_MATRIX is overflowing with a single push!
    # glPopMatrix();
    glLoadMatrixd( matrix ) # should have un-decorated alias for this...
    
    glMatrixMode(GL_MODELVIEW);

def display():
    glutSetWindow(window);
    glClearColor (0.0, 0.0, (time.time()%1.0)/1.0, 0.0)
    glClear (GL_COLOR_BUFFER_BIT)
    drawText( 'hello', 20,20, size[0],size[1] )
    #glutBitmapCharacter( GLUT_BITMAP_8_BY_13, ord('a'))
    glutSolidTeapot( .2 )
    glFlush ()
    glutSwapBuffers()

size = (250,250)

def reshape( *args ):
    global size 
    size = args
    glViewport( *( (0,0)+args) )
    display()

def ontimer( *args ):
    print(('timer', args, '@time', time.time()-start))
    glutTimerFunc( 1000, ontimer, 24 )

def idle():
    delta = time.time()-start
    if delta < 10:
        global size 
        x,y = size 
        if delta < 5:
            change = +1
        else:
            change = -1
        x = x-change
        y = y+change
        if x < 1:
            x = 1
        if y < 1:
            y = 1
        glutReshapeWindow( x, y )
        size = (x,y)
        glutSetWindow(window)
        glutPostRedisplay()
    else:
        glutDestroyWindow( window )
        print('window destroyed')
        try:
            if fgDeinitialize: fgDeinitialize(False)
        except NameError:
            pass # Older PyOpenGL, you may see a seg-fault here...
        import sys
        sys.exit( 0 )

def printFunction( name ):
    def onevent( *args ):
        print(('%s -> %s'%(name, ", ".join( [str(a) for a in args ]))))
    return onevent



if __name__ == "__main__":
    import sys
    newArgv = glutInit(sys.argv)
    print(('newArguments', newArgv))
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB )
    glutInitWindowSize(250, 250)
    glutInitWindowPosition(100, 100)
    window = glutCreateWindow(as_8_bit("hello"))
    print(('window', repr(window)))
    glutDisplayFunc(display)
    glutReshapeFunc(reshape)
    glutMouseFunc(printFunction( 'Mouse' ))
    glutEntryFunc(printFunction( 'Entry' ))
    glutKeyboardFunc( printFunction( 'Keyboard' ))
    glutKeyboardUpFunc( printFunction( 'KeyboardUp' ))
    glutMotionFunc( printFunction( 'Motion' ))
    glutPassiveMotionFunc( printFunction( 'PassiveMotion' ))
    glutVisibilityFunc( printFunction( 'Visibility' ))
    glutWindowStatusFunc( printFunction( 'WindowStatus' ))
    glutSpecialFunc( printFunction( 'Special' ))
    glutSpecialUpFunc( printFunction( 'SpecialUp' ))
    glutTimerFunc( 1000, ontimer, 23 )
    
    glutIdleFunc( idle )
    glutMainLoop()



"""Test GLUT forward-compatible mode..."""
from __future__ import print_function
import OpenGL
OpenGL.FORWARD_COMPATIBLE_ONLY = True
OpenGL.ERROR_CHECKING = True
#OpenGL.USE_ACCELERATE = False
from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *
#from OpenGL.GL import shaders
import time
start = time.time()

window = None

def display():
    try:
        glutSetWindow(window);
        glClearColor (0.0, 0.0, (time.time()%1.0)/1.0, 0.0)
        glClear (GL_COLOR_BUFFER_BIT)
        try:
            glGetString( GL_EXTENSIONS )
        except GLError:
            pass 
        else:
            print('Egads, glGetString should not have worked!')
        assert bool( glGenVertexArrays ), "Should have vertex array support in 3.2"
        glFlush ()
        glutSwapBuffers()
    except Exception:
        glutDestroyWindow( window )
        raise

size = (250,250)

def reshape( *args ):
    global size 
    size = args
    glViewport( *( (0,0)+args) )
    display()

def ontimer( *args ):
#    print 'timer', args, '@time', time.time()-start
    glutTimerFunc( 1000, ontimer, 24 )

def idle():
    delta = time.time()-start
    if delta < 10:
        global size 
        x,y = size 
        if delta < 5:
            change = +1
        else:
            change = -1
        x = x-change
        y = y+change
        if x < 1:
            x = 1
        if y < 1:
            y = 1
        glutReshapeWindow( x, y )
        size = (x,y)
        glutSetWindow(window)
        glutPostRedisplay()
    else:
        glutDestroyWindow( window )
        print('window destroyed')
        try:
            if fgDeinitialize: fgDeinitialize(False)
        except NameError:
            pass # Older PyOpenGL, you may see a seg-fault here...
        import sys
        sys.exit( 0 )

def printFunction( name ):
    def onevent( *args ):
        print('%s -> %s'%(name, ", ".join( [str(a) for a in args ])))
    return onevent



if __name__ == "__main__":
    import sys
    newArgv = glutInit(sys.argv)
    glutInitContextVersion(3, 1)
    glutInitContextFlags(GLUT_FORWARD_COMPATIBLE)
    glutInitContextProfile(GLUT_CORE_PROFILE)
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH )

    glutSetOption(
        GLUT_ACTION_ON_WINDOW_CLOSE,
        GLUT_ACTION_GLUTMAINLOOP_RETURNS
    );
    glutInitWindowSize(250, 250)
    glutInitWindowPosition(100, 100)
    window = glutCreateWindow("hello")
    print('window', repr(window))
    glutDisplayFunc(display)
    glutReshapeFunc(reshape)
    glutMouseFunc(printFunction( 'Mouse' ))
    glutEntryFunc(printFunction( 'Entry' ))
    glutKeyboardFunc( printFunction( 'Keyboard' ))
    glutKeyboardUpFunc( printFunction( 'KeyboardUp' ))
    glutMotionFunc( printFunction( 'Motion' ))
    glutPassiveMotionFunc( printFunction( 'PassiveMotion' ))
    glutVisibilityFunc( printFunction( 'Visibility' ))
    glutWindowStatusFunc( printFunction( 'WindowStatus' ))
    glutSpecialFunc( printFunction( 'Special' ))
    glutSpecialUpFunc( printFunction( 'SpecialUp' ))
    glutTimerFunc( 1000, ontimer, 23 )
    
    glutIdleFunc( idle )
    glutMainLoop()



import glfw
from OpenGL.GL import *
if not glfw.init():
    sys.exit()
glfw.window_hint(glfw.SAMPLES, 4)
glfw.window_hint(glfw.CONTEXT_VERSION_MAJOR, 1)
glfw.window_hint(glfw.CONTEXT_VERSION_MINOR, 4)
window = glfw.create_window(800, 600, "Hello World", None, None)
if not window:
    sys.exit()
glfw.make_context_current(window)
glEnable(GL_BLEND)
# glDisable(GL_BLEND)
glClearColor(1.0/255.0*68.0, 1.0/255.0*68.0, 1.0/255.0*68.0, 1.0)
glClear(GL_COLOR_BUFFER_BIT)
glViewport(0, 0, 800, 600)
glMatrixMode(GL_PROJECTION)
glLoadIdentity()
glOrtho(0.0, 800.0, 600.0, 0.0, 0.0, 1.0)
 
def drawOneLine(x1, y1, x2, y2, width):
    glDisable(GL_LINE_SMOOTH);
    glLineWidth(width)
    glBegin(GL_LINES)
    glVertex2f(x1, y1)
    glVertex2f(x2, y2)
    glEnd()
   
if __name__ == "__main__":
    while not glfw.window_should_close(window):
        glfw.poll_events()            
        glClear(GL_COLOR_BUFFER_BIT)
        for y in range(1,20):
            drawOneLine(10,20*y+.5,100,20*y+.5,y*.5)
        glfw.swap_buffers(window)
    glfw.destroy_window(window)
    glfw.terminate()



#! /usr/bin/env python
"""Registry for loading Khronos API definitions from XML files"""
from lxml import etree as ET
import os, sys, json, logging
from OpenGL._bytes import as_8_bit, unicode, as_str
log = logging.getLogger( __name__ )
HERE = os.path.dirname( __file__ )

LENGTH_OVERRIDES={
    'glGetPolygonStipple': {
        'mask': str(32*32/8), # 32x32 bits
    },
    'glGetUniformfv': {
        'params': None,
    },
    'glGetUniformiv': {
        'params': None,
    },
#    'glShaderSourceARB': {
#        'string': None,
#    },
#    'glShaderSource': {
#        'string': None,
#    },
}


class Registry( object ):
    def __init__( self ):
        self.type_set = {}
        self.enum_namespaces = {}
        self.enum_groups = {}
        self.enumeration_set = {}
        self.command_set = {}
        self.apis = {}
        self.feature_set = {}
        self.extension_set = {}
        self.output_mapping = json.loads( open( os.path.join( HERE, 'gl_out_parameters.json' )).read())
        self.output_enum_groups = {}
        
    def load( self, tree ):
        """Load an lxml.etree structure into our internal descriptions"""
        self.dispatch( tree, None )
    
    def dispatch( self, tree, context=None):
        """Dispatch for all children of the element"""
        for element in tree:
            if isinstance( element.tag, (str,unicode)):
                method = getattr( self, element.tag, None )
                if method:
                    method( element, context )
                else:
                    print('Expand', element.tag)
                    self.dispatch( element, context )
    
    def type( self, element, context=None ):
        name = element.get('name')
        if not name:
            name = element.find('name').text 
        self.type_set[as_str(name)] = element 
    
    def debug_types( self ):
        for name,type in self.types.items():
            print(name, type)
    
    def enums( self, element, context=None ):
        name = as_str(element.get('namespace'))
        if name not in self.enum_namespaces:
            namespace = EnumNamespace(name)
            self.enum_namespaces[name] = namespace
        else:
            namespace = self.enum_namespaces[name]
        self.dispatch( element, namespace )
    
    def enum( self, element, context=None ):
        if isinstance( context, EnumNamespace ):
            name,value = as_str(element.get('name')),element.get('value')
            enum = Enum( name, value )
            context.append( enum )
            self.enumeration_set[name] = enum
        elif isinstance( context, (Require,Remove)):
            context.append( self.enumeration_set[element.get('name')] )
        elif isinstance( context, EnumGroup ):
            name = element.get('name')
            assert name, 'No name on %s'%ET.tostring(element)
            context.append( as_str(name) )
    
    def debug_enums( self ):
        for name,namespace in self.enum_namespaces.items():
            print('Namespace', namespace.namespace)
            for enum in namespace:
                print('  ', enum)
    
    def command( self, element, context=None ):
        """Parse command definition into structured format"""
        proto = element.find( 'proto' )
        if proto is not None:
            name = as_str(proto.find('name').text)
            assert name, 'No name in command: %s'%(ET.tostring( element))
            return_type = self._type_decl( proto )
            assert return_type, 'No return type in command: %s'%(ET.tostring( element))
            arg_names = []
            arg_types = []
            lengths = {}
            groups = {}
            for param in [x for x in element if x.tag == 'param']:
                pname = as_str(param.find( 'name' ).text)
                arg_names.append( pname )
                arg_types.append( self._type_decl( param ))
                if param.get( 'len' ):
                    length = param.get('len')
                    while length.endswith( '*1' ):
                        length = length[:-2]
                    length = LENGTH_OVERRIDES.get(name,{}).get(pname,length)
                    lengths[pname] = length 
                if param.get( 'group' ):
                    groups[pname] = param.get('group')
            aliases = []
            for alias in [x for x in element if x.tag == 'alias']:
                aliases.append( alias.get('name') )
            # Process lengths to look for output parameters
            outputs = self.output_mapping.get( name )
            command = Command( name, return_type, arg_names, arg_types, aliases, lengths,groups, outputs=outputs )
            self.command_set[name] = command
        elif isinstance( context, (Require,Remove)):
            context.append( self.command_set[element.get('name')])
    
    def _type_decl( self, proto ):
        """Get the string type declaration for parent (proto/param)"""
        return_type = []
        if proto.text:
            return_type.append( proto.text )
        for item in proto:
            if item.tag == 'name':
                break
            else:
                if item.text:
                    return_type.append(item.text.strip())
                if item.tail:
                    return_type.append(item.tail.strip())
        return ' '.join( [x for x in return_type if x] ) or 'void'
    
    def debug_commands( self ):
        for name,command in sorted(self.command_set.items()):
            print( command )
        
    def feature( self, element, context=None ):
        api,name,number = [element.get(x) for x in ('api','name','number')]
        feature = Feature( api, name, number )
        self.feature_set[name] = feature 
        self.dispatch( element, feature )
    def extension( self, element, context=None ):
        name,apis,require = [element.get(x) for x in ['name','supported','protect']]
        extension = Extension( name, apis.split('|'),require)
        self.extension_set[name] = extension
        self.dispatch( element, extension )
    def unused( self, element, context=None):
        pass
    def group( self, element, context=None):
        name = element.get('name')
        current = self.enum_groups.get( name )
        if current is None:
            current = self.enum_groups[name] = EnumGroup( name )
        self.dispatch( element, current )
    
    def require( self, element, context ):
        if isinstance( context, (Feature,Extension)):
            profile,comment = element.get('profile'),element.get('comment')
            require = Require( profile, comment )
            context.append( require )
            self.dispatch( element, require )
    def remove( self, element, context ):
        if isinstance( context, Feature):
            profile,comment = element.get('profile'),element.get('comment')
            remove = Remove( profile, comment )
            context.append( remove )
            self.dispatch( element, remove )
    
    def debug_apis( self ):
        print( [x.api for x in self.feature_set.values()])

class EnumNamespace( list ):
    def __init__( self, namespace, *args ):
        self.namespace = namespace 
        super( EnumNamespace, self ).__init__(*args)
class EnumGroup( list ):
    def __init__( self, name, *args ):
        self.name = name 
        super( EnumGroup, self ).__init__( *args )
class Enum( object ):
    def __init__( self, name, value ):
        self.name = name 
        self.value = value
    def __repr__( self ):
        return '%s = %s'%( self.name, self.value )

class Command( object ):
    def __init__( self, name, returnType, argNames, argTypes, aliases=None, lengths=None, groups=None, outputs=None ):
        self.name =name 
        self.returnType = returnType 
        self.argNames = argNames 
        self.argTypes = argTypes
        self.aliases = aliases or []
        self.lengths = lengths or {}
        self.groups = groups or {}
        self.outputs = outputs or {}
        self.output_groups = {}
        self.size_dependencies = self.calculate_sizes()
    def __repr__( self ):
        return '%s %s( %s )'%( 
            self.returnType, 
            self.name, 
            ', '.join([
                '%s %s'%(typ,name) 
                for (typ,name) in zip( self.argTypes,self.argNames )
            ])
        )
    def calculate_sizes( self ):
        result = []
        other_lengths = self.lengths.copy()
        for target in self.outputs.keys():
            definition = self.lengths.get( target )
            if definition is None:
                if target not in self.argNames:
                    # may be a discrepency between .spec and xml registry file...
                    if target == 'params' and 'data' in self.argNames:
                        target = 'data'
                        definition = self.lengths.get( 'data' )
            if target in other_lengths:
                del other_lengths[target]
        
            if definition is None:
                result.append( (target, Output() ))
            elif definition.startswith( 'COMPSIZE' ):
                variables = definition[8:].strip('()').split(',')
                output_groups = {}
                if len(variables) == 1:
                    # for now we only support automated single-dependency wrapping...
                    for var in variables:
                        if var in self.groups:
                            output_groups.setdefault( self.groups[var], []).append( target )
                    self.output_groups.update( output_groups )
                result.append( (target,Compsize( variables, output_groups )))
            elif definition.isdigit():
                result.append( (target,Staticsize( int(definition,10))))
            elif '*' in definition:
                var,multiple = definition.split('*')
                result.append( (target,Multiple( var, int(multiple,10))))
            else:
                result.append( (target,Dynamicsize( definition )))
        for (target,length) in other_lengths.items():
            if length is None:
                # length/array-conversion suppressed
                continue 
            if length.isdigit():
                result.append( (target,StaticInput( int(length,10))))
            elif length.startswith( 'COMPSIZE' ):
                result.append( (target,Input(length[9:-1])))
            elif length in self.argNames:
                result.append( (target,DynamicInput(length)))
            elif '*' in length:
                params = length.split('*')
                in_set = [
                    x for x in params 
                    if x in self.argNames
                ]
#                if params == in_set:
                result.append( (target,MultipleInput(params)))
#                else:
#                    result.append( (target, Input()))
            else:
                raise RuntimeError( (target,length))
        return dict(result)
class IsInput(object):
    pass
class Input( IsInput, object ):
    """Unsized Input Parameter"""
    def __init__( self, value=None ):
        self.value = value
    def __repr__( self ):
        return repr(self.value)
class StaticInput( IsInput,int ):
    """Statically sized input parameter"""
class DynamicInput( IsInput,str ):
    """Dynamically sized based on other parameter"""
class MultipleInput( IsInput,list ):
    """Size depends on multiple elements being multiplied"""
    def __str__( self ):
        return '*'.join( self )
        
class Output( object ):
    """Unsized output parameter"""
class Compsize( list ):
    """Compute size based on other variables"""
    def __init__( self, iterable, groups=None ):
        super( Compsize,self ).__init__( iterable )
        self.groups = groups

class Staticsize( int ):
    """Static output array size"""
class Dynamicsize( str ):
    """Sized by the value in dynamic variable"""
class Multiple( list ):
    """Variable * static size for array"""

# The order-dependent set of require/remove holding features/extensions
class Module( list ):
    """Base class for Features and Extensions"""
    feature = False
    def __init__( self, name ):
        self.name = name 
    def members( self, of_type=None ):
        members = []
        for req in self:
            if req.require:
                for item in req:
                    if of_type is not None:
                        if isinstance( item, of_type ):
                            members.append( item )
                    else:
                        members.append( item )
        return members
    def enums( self ):
        return self.members( Enum )
    def commands( self ):
        return self.members( Command )


class Feature( Module ):
    feature = True
    NORMALIZERS = {
        'GL_VERSION_ES_CM_1_0': 'GLES_VERSION_1_0',
        'GL_ES_VERSION_2_0': 'GLES_VERSION_2_0',
        'GL_ES_VERSION_3_0': 'GLES_VERSION_3_0',
    }
    def __init__( self, api,name,number ):
        super( Feature, self ).__init__(self.NORMALIZERS.get(name,name))
        self.api = api 
        if name == 'GL_ES_VERSION_3_0':
            self.api = 'gles3'
        self.number = number 
    _profiles = None
    @property 
    def profiles( self ):
        """Create set of profiles with subsets of our functionality"""
        if self._profiles is None:
            profiles = {}
            for req in self:
                # Logic isn't right here, there's a base and then 
                # a set of profiles which customize the base...
                profile = req.profile or ''
                set = profiles.get( profile )
                if set is None:
                    set = Module( profile or '' )
                    set.feature = True
                    profiles[profile] = set
                if req.require:
                    
                    set.extend( req )
                else:
                    for item in req:
                        while item in set:
                            set.remove( item )
            self._profiles = sorted(profiles.values(),key=lambda x: x.name)
        return self._profiles
    
class Extension( Module ):
    def __init__(self, name, apis, require=None ):
        super( Extension, self ).__init__(name)
        self.apis = apis # only available for these APIs
        self.require = require
    @property
    def profiles( self ):
        module = Module( 'default' )
        module.extend( self )
        return module
    
class Require( list ):
    require = True
    remove = False
    def __init__( self, profile=None, comment=None ):
        self.profile = profile 
        self.comment = comment 
        super( Require, self ).__init__()
class Remove( list ):
    require = False
    remove = True
    def __init__( self, profile=None, comment=None ):
        self.profile = profile 
        self.comment = comment 
        super( Remove, self ).__init__()

def parse( xmlfile ):
    registry = Registry()
    registry.load( ET.fromstring( open( xmlfile, 'rb' ).read()) )
    return registry 


if __name__ == "__main__":
    if sys.argv[1:]:
        for file in sys.argv[1:]:
            print(file)
            registry = parse( file )
    
    #registry.debug_types()
    #registry.debug_enums()
    #registry.debug_commands()
    #registry.debug_apis()
    



#! /usr/bin/env python
from __future__ import print_function
import egltest
from numpy import array
from OpenGL.arrays.vbo import VBO
from OpenGL.arrays import ArrayDatatype
from OpenGL.GLES2 import *
from OpenGL.GLES2 import shaders

@egltest.egltest( api='es2' )
def test_gl( ):
    glClearColor( 1,0,0, 0 )
    glClear( GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT )
    
    print('Error before compilation?', glGetError())
    shader = shaders.compileProgram(
        shaders.compileShader( '''#version 130
    attribute vec3 position;
    void main() {
        gl_Position = vec4( position, 0 );
    }''', GL_VERTEX_SHADER),
        shaders.compileShader( '''#version 130
    void main() {
        gl_FragColor = vec4( 0,1,0,.5 );
    }''', GL_FRAGMENT_SHADER),
    )
    vbo = VBO( array([
        (0,1,0),
        (1,-1,0),
        (-1,-1,0),
    ],dtype='f'))
    position_location = glGetAttribLocation(
        shader, 'position'
    )
    stride = 3*4
    with vbo:
        with shader:
            glEnableVertexAttribArray( position_location )
            stride = 3*4
            glVertexAttribPointer(
                position_location,
                3, GL_FLOAT,False, stride, vbo
            )
            glDrawArrays( GL_TRIANGLES, 0, 3 )

    
if __name__ == "__main__":
    test_gl()
    



#! /usr/bin/env python
"""Get the EGL and ES header files"""
import requests, os, logging, re
from funcparse import Function,Helper
log = logging.getLogger( 'build_gles' )
HERE = os.path.dirname( __file__ )
HEADER_DIR = os.path.join( HERE, 'es','include' )
EGL_DIR = os.path.join( HERE, '..','OpenGL','egl' )

HEADER_FILES = [
    # EGL 1.4...
    'http://www.khronos.org/registry/egl/api/KHR/khrplatform.h',
    'http://www.khronos.org/registry/egl/api/EGL/egl.h',
    'http://www.khronos.org/registry/egl/api/EGL/eglext.h',
    'http://www.khronos.org/registry/egl/api/EGL/eglplatform.h',
    # ES 3
    'http://www.khronos.org/registry/gles/api/3.0/gl3.h',
    'http://www.khronos.org/registry/gles/api/3.0/gl3ext.h',
    'http://www.khronos.org/registry/gles/api/3.0/gl3platform.h',
    # ES 2
    'http://www.khronos.org/registry/gles/api/2.0/gl2.h',
    'http://www.khronos.org/registry/gles/api/2.0/gl2ext.h',
    'http://www.khronos.org/registry/gles/api/2.0/gl2platform.h',
]

def ensure_headers( force=False ):
    if not os.path.exists( HEADER_DIR ):
        os.makedirs( HEADER_DIR )
    for url in HEADER_FILES:
        expected = os.path.join( HEADER_DIR, url.split('/')[-1] )
        if force or not os.path.exists( expected ):
            log.info( 'Downloading %s into %s', url, expected )
            response = requests.get( url )
            if response.status_code == 200:
                open( expected, 'w' ).write( response.content )
                log.info( '  Download complete: %s bytes', len(response.content) )
            else:
                log.error( '  Download failed, status code %s: %s', response.status_code, response.content )
                raise RuntimeError( url )
        

CONSTANT_DEFINE = re.compile( r'^#define\W+(?P<name>\w+)\W+(?P<value>(0x[0-9a-fA-F]+)|([0-9]+))', re.M )
def find_constants( header ):
    """Given a header, try to find our constants"""
    for constant in CONSTANT_DEFINE.finditer( header ):
        yield constant.groupdict()

def parse_args( args ):
    args = args.strip()
    if args == 'void':
        return []
    else:
        return [parse_arg( a ) for a in args.split(',')]

ARG_PARSE = re.compile( r'^\W*(?P<type>.+?)[ \t]*(?P<name>[a-zA-Z0-9_]+)$', re.DOTALL|re.M)
def parse_arg( arg ):
    arg = arg.strip()
    try:
        arg = ARG_PARSE.search( arg ).groupdict()
    except Exception as err:
        err.args += ( arg, )
        raise
    else:
        arg['type'] = parse_type( arg['type'] )
        return arg

def parse_type( typ ):
    typ = typ.strip()
    if typ.startswith( 'const' ):
        typ = typ[5:].strip()
    indirections = 0
    while typ.endswith( '*' ):
        typ = typ[:-1].strip()
        indirections += 1
    while typ.startswith( 'struct' ):
        typ = typ[6:].strip()
    
    for i in range( indirections ):
        typ = 'pointer(%s)'%( typ, )
    return typ

EGL_PREFIX = '''"""egl wrapper for PyOpenGL"""
# THIS FILE IS AUTO-GENERATED DO NOT EDIT!
from OpenGL import platform as _p
from OpenGL.khr import *
from OpenGL import khr as _cs
from OpenGL import arrays
from OpenGL.constant import IntConstant as _C
import ctypes

# Callback types, this is a hack to avoid making the 
# khr module depend on the platform or needing to change generator for now...
CALLBACK_TYPE = _p.PLATFORM.functionTypeFor( _p.PLATFORM.EGL )
_cs.EGLSetBlobFuncANDROID = CALLBACK_TYPE( ctypes.c_voidp, EGLsizeiANDROID, ctypes.c_voidp, EGLsizeiANDROID )
_cs.EGLGetBlobFuncANDROID = CALLBACK_TYPE( ctypes.c_voidp, EGLsizeiANDROID, ctypes.c_voidp, EGLsizeiANDROID )

EGL_DEFAULT_DISPLAY = EGLNativeDisplayType()
EGL_NO_CONTEXT = EGLContext()
EGL_NO_DISPLAY = EGLDisplay()
EGL_NO_SURFACE = EGLSurface()
EGL_DONT_CARE = -1

def _f( function ):
    return _p.createFunction( function,_p.PLATFORM.EGL,None,False)
'''
EGL_SUFFIX = '''
del ctypes 
del arrays 
'''
    
EGL_API = re.compile( r'^(EGLAPI|GL_APICALL)[ \t\n](?P<returntype>[a-zA-Z0-9_ *]+)\W+(EGLAPIENTRY|GL_APIENTRY)\W+(?P<name>[_a-z0-9A-Z]+)[ \t]*[(](?P<arguments>[^)]+)[)]', re.M|re.DOTALL )
EGL_SUPPRESS = set([
    'eglGetProcAddress',
])

def load_header( header ):
    filename = os.path.join( HEADER_DIR, header )
    if not os.path.exists( filename ):
        raise RuntimeError( "Missing header %s", filename )
    return open( filename ).read()

def write_module( constant_headers, headers, module, suppress=None, prefix=None, suffix=None ):
    suppress = suppress or set()
    
    constant_headers = [load_header(h) for h in constant_headers]
    headers = [load_header(h) for h in headers ]
    
    constants = []
    for header in constant_headers:
        constants.extend( find_constants( header ))
    functions = []
    for header in headers:
        for function in EGL_API.finditer( header ):
            function = function.groupdict()
            #function['returntype'] = parse_type( function['returntype'] )
            function = Function( function['returntype'], function['name'], function['arguments'] )
            if function.name not in suppress:
                functions.append( function )
    if not os.path.exists( os.path.dirname( module ) ):
        os.makedirs( os.path.dirname( module ) )
    with open( module, 'w' ) as fh:
        fh.write(prefix or '')
        for constant in constants:
            fh.write( '%(name)s=_C(%(name)r,%(value)s)\n'%constant )
        for function in functions:
            fh.write( function.declaration() )
            fh.write('\n')
            log.debug( '%s', function['name'] )
        fh.write(suffix or '')
    log.info( 'File: %s', module )
    log.info( '  %s constants', len(constants))
    log.info( '  %s functions', len(functions))

def egl():
    module = os.path.join( EGL_DIR, '__init__.py' )
    log.info( 'Starting EGL module %s', module )
    khr = 'khrplatform.h'
    platform = 'eglplatform.h'
    core = 'egl.h'
    ext = 'eglext.h'
    mesa= 'eglmesaext.h'
    
    write_module( 
        constant_headers = (khr,platform,core,ext,mesa),
        headers= (core,ext,mesa),
        module = module,
        suppress = EGL_SUPPRESS,
        prefix = EGL_PREFIX,
        suffix= EGL_SUFFIX,
    )

GLES_PREFIX = '''"""GLES %(es_version)s wrapper for PyOpenGL"""
from OpenGL import constants as _cs 
from OpenGL import platform as _p
from OpenGL import arrays
from OpenGL.constant import IntConstant as _C
import ctypes

def _f( function ):
    return _p.createFunction( function,_p.GL,None,False)
'''
GLES_SUFFIX = '''
'''
GLES_SUPPRESS = set([
#    'glRenderbufferStorageMultisampleAPPLE', # missing parameter names...
#    'glRenderbufferStorageMultisampleEXT',
#    'glFramebufferTexture2DMultisampleEXT',
])

def gles():
    """Create the GLES 2 and 3 modules"""
    for es_version in (2,3):
        target_dir = os.path.join( HERE, '..', 'OpenGL','es%(es_version)s'%locals() )
        target_module = os.path.join( target_dir, '__init__.py' )
        headers = [ h%locals() for h in [
            'gl%(es_version)s.h','gl%(es_version)sext.h','gl%(es_version)splatform.h']
        ]
        write_module( 
            constant_headers = headers,
            headers= headers,
            module = target_module,
            suppress = GLES_SUPPRESS,
            prefix = GLES_PREFIX%locals(),
            suffix= GLES_SUFFIX,
        )
        
def main():
    ensure_headers()
    egl()
    gles()

if __name__ == "__main__":
    logging.basicConfig( level=logging.INFO )
    main()
